/*
 * Sean Landman
 * Data Mining for Biomedical Informatics Group
 * University of Minnesota - Twin Cities
 */

package edu.umn.cs.kumarbio.shear.aligners;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import edu.umn.cs.kumarbio.shear.ShearRuntimeException;
import edu.umn.cs.kumarbio.shear.ShearSV;
import edu.umn.cs.kumarbio.shear.system.RunSystemCommand;

/**
 * Represents a program that is used to do local alignment with BWA. This does
 * single-end read alignment of one FASTQ file to a specified reference
 * sequence. The BWA index files can either be pre-generated or created in the
 * constructor. Implementations of this class will specify particular types of
 * BWA local alignment to do (i.e. BWA-MEM vs. BWA-SW).
 */
public abstract class BwaAligner {

  // The path to the reference file to align to
  private String  faFilePath;

  // Whether or not to generate BWA index files for this aligner
  private boolean generateBwaIndex;

  // The instance of ShearSV that this is called from
  private ShearSV shearSvInstance;

  /**
   * Creates a new BwaMemAligner object for the specified reference file. The
   * generateBwaIndex parameter can be set to true to generate BWA index files
   * for this aligner instance to use. If generateBwaIndex is false, then the
   * proper BWA index files are expected to be present already with the
   * reference file path being the prefix for the BWA index files.
   *
   * @param faFile
   *          The path to the reference file to align to
   * @param generateBwaIndex
   *          Whether or not to generate BWA index files for this aligner
   * @param shearSvInstance
   *          The instance of ShearSV that this is called from
   */
  public BwaAligner(String faFilePath, boolean generateBwaIndex, ShearSV shearSvInstance) {

    // Set instance variables
    this.faFilePath = faFilePath;
    this.generateBwaIndex = generateBwaIndex;
    this.shearSvInstance = shearSvInstance;

    // Generate BWA index if necessary
    if (generateBwaIndex) {
      RunSystemCommand.run(
          new ProcessBuilder("bwa", "index", faFilePath),
          System.out,
          "BWA",
          shearSvInstance.isDebugMode());
    }

    // Check for existence of BWA index
    if (!new File(faFilePath + ".bwt").exists()) {
      throw new ShearRuntimeException("Required BWA index file does not exist: " + faFilePath + ".bwt");
    }

  }

  /**
   * Returns the path to the reference file to align to.
   *
   * @return The path to the reference file to align to
   */
  public String getFaFilePath() {
    return faFilePath;
  }

  /**
   * Returns the ShearSV instance that this is being called from.
   *
   * @return The ShearSV instance that this is being called from
   */
  public ShearSV getShearSvInstance() {
    return shearSvInstance;
  }

  /**
   * Returns true if debugging mode is active, false otherwise.
   *
   * @return True if debugging mode is active, false otherwise
   */
  public boolean isDebugMode() {
    return getShearSvInstance().isDebugMode();
  }

  /**
   * Performs a local alignment using this instance of a BWA local aligner on
   * the specified FASTQ file and outputting to the specified SAM file location.
   *
   * @param fastqInputFilePath
   *          Path to the FASTQ file to use as input
   * @param samOutputFilePath
   *          Path to the SAM file to output to
   * @throws FileNotFoundException
   *           If there is a problem creating the SAM output file to write to
   * @throws IOException
   *           If there is a problem closing the SAM output file
   */
  public void align(String fastqInputFilePath, String samOutputFilePath) throws FileNotFoundException, IOException {
    align(fastqInputFilePath, samOutputFilePath, new ArrayList<String>());
  }

  /**
   * Performs a local alignment using this instance of a BWA local aligner on
   * the specified FASTQ file and outputting to the specified SAM file location.
   *
   * @param fastqInputFilePath
   *          Path to the FASTQ file to use as input
   * @param samOutputFilePath
   *          Path to the SAM file to output to
   * @param options
   *          The additional options to use while running this aligner
   * @throws FileNotFoundException
   *           If there is a problem creating the SAM output file to write to
   * @throws IOException
   *           If there is a problem closing the SAM output file
   */
  public abstract void align(String fastqInputFilePath, String samOutputFilePath, List<String> options)
      throws FileNotFoundException, IOException;

  /**
   * Closes this instance of a BWA local aligner by deleting all BWA index files
   * if they were generated just for this instance of a BWA local aligner.
   */
  public void close() {
    if (generateBwaIndex) {
      new File(faFilePath + ".amb").delete();
      new File(faFilePath + ".ann").delete();
      new File(faFilePath + ".bwt").delete();
      new File(faFilePath + ".pac").delete();
      new File(faFilePath + ".sa").delete();
    }
  }

}
