/*
 * Sean Landman
 * Data Mining for Biomedical Informatics Group
 * University of Minnesota - Twin Cities
 */

package edu.umn.cs.kumarbio.shear.heterogeneity;

import edu.umn.cs.kumarbio.shear.ShearRuntimeException;
import edu.umn.cs.kumarbio.shear.variants.Location;

/**
 * This represents a breakpoint for a variant.
 */
public class Breakpoint implements Comparable<Breakpoint> {

  // The 5'-most possible location for this breakpoint (i.e. coordinate just before the break)
  private Location        fivePrimeLoc;

  // The 3'-most possible location for this breakpoint (i.e. coordinate just before the break)
  private Location        threePrimeLoc;

  // The information about counts of reads that are clipped at or span the breakpoint
  private BreakpointReads bpReads;

  /**
   * Creates a new Breakpoint object using the specified location range. The
   * actual breakpoint could occur anywhere between the two specified locations
   * due to microhomology. The location represents the coordinate just before
   * the break. Initializes the breakpoint reads to zeros.
   *
   * @param fivePrimeLoc
   *          The 5'-most possible location for this breakpoint (i.e. coordinate
   *          just before the break)
   * @param threePrimeLoc
   *          The 3'-most possible location for this breakpoint (i.e. coordinate
   *          just before the break)
   */
  public Breakpoint(Location fivePrimeLoc, Location threePrimeLoc) {

    // Validate arguments
    if (fivePrimeLoc == null || threePrimeLoc == null) {
      throw new ShearRuntimeException("Null location(s) specified for breakpoint");
    } else if (!fivePrimeLoc.getChr().equals(threePrimeLoc.getChr())) {
      throw new ShearRuntimeException("Possible locations for breakpoint must be on same chromosome");
    } else if (fivePrimeLoc.getCoord() > threePrimeLoc.getCoord()) {
      throw new ShearRuntimeException("3'-most possible location must be at least as high as 5'-most possible location");
    }

    // Set instance variables
    this.fivePrimeLoc = fivePrimeLoc;
    this.threePrimeLoc = threePrimeLoc;
    bpReads = new BreakpointReads(0, 0, 0, 0);

  }

  /**
   * Returns the 5'-most possible location for this breakpoint (i.e. coordinate
   * just before the break).
   *
   * @return The 5'-most possible location for this breakpoint (i.e. coordinate
   *         just before the break)
   */
  public Location getFivePrimeLoc() {
    return fivePrimeLoc;
  }

  /**
   * Returns the 3'-most possible location for this breakpoint (i.e. coordinate
   * just before the break).
   *
   * @return The 3'-most possible location for this breakpoint (i.e. coordinate
   *         just before the break)
   */
  public Location getThreePrimeLoc() {
    return threePrimeLoc;
  }

  /**
   * Returns the information about counts of reads that are clipped at or span
   * the breakpoint.
   *
   * @return The information about counts of reads that are clipped at or span
   *         the breakpoint
   */
  public BreakpointReads getBpReads() {
    return bpReads;
  }

  /**
   * Returns true if the two breakpoints have possible locations that overlap
   * one another, or that are within the allowable separation to one another.
   * Use '0' for allowableSeparation to check if the two breakpoints overlap.
   *
   * @param other
   *          The breakpoint to compare to
   * @param allowableSeparation
   *          The maximum number of bases that the two breakpoint ranges can be
   *          separated by in order to return true.
   * @return True if the two breakpoints have possible locations that overlap
   *         one another, or that are within the allowable separation to one
   *         another
   */
  public boolean overlapsOrAdjacentTo(Breakpoint other, int allowableSeparation) {
    if (!getFivePrimeLoc().getChr().equals(other.getFivePrimeLoc().getChr())) {
      return false;
    } else {
      return Math.max(getFivePrimeLoc().getCoord(), other.getFivePrimeLoc().getCoord()) <= Math.min(getThreePrimeLoc()
          .getCoord(), other.getThreePrimeLoc().getCoord())
          + allowableSeparation;
    }
  }

  @Override
  public int compareTo(Breakpoint other) {

    // Return 0 if breakpoints are equal
    if (equals(other)) {
      return 0;
    }

    // Return whichever has first 5' location
    if (getFivePrimeLoc().compareTo(other.getFivePrimeLoc()) < 0) {
      return -1;
    } else if (getFivePrimeLoc().compareTo(other.getFivePrimeLoc()) > 0) {
      return +1;
    }

    // Return whichever has first 3' location
    if (getThreePrimeLoc().compareTo(other.getThreePrimeLoc()) < 0) {
      return -1;
    } else if (getThreePrimeLoc().compareTo(other.getThreePrimeLoc()) > 0) {
      return +1;
    }

    // If locations are the same, compare the strings of the BreakpointReads
    if (getBpReads().toString().compareTo(other.getBpReads().toString()) < 0) {
      return -1;
    } else if (getBpReads().toString().compareTo(other.getBpReads().toString()) > 0) {
      return +1;

      // Throw error for exact same BreakpointReads string (should be unreachable code)
    } else {
      throw new ShearRuntimeException("Unexpected behavior while comparing order of Breakpoint objects"
          + " (Same locations and breakpoint reads, but objects not equal)");
    }

  }

  @Override
  public boolean equals(Object obj) {
    if (obj == null || !(obj instanceof Breakpoint)) {
      return false;
    } else {
      Breakpoint other = (Breakpoint) obj;
      return getFivePrimeLoc().equals(other.getFivePrimeLoc()) && getThreePrimeLoc().equals(other.getThreePrimeLoc())
          & getBpReads().equals(other.getBpReads());
    }
  }

  @Override
  public String toString() {
    return String.format("%s (%dbp microhomology) (reads: %s)", getFivePrimeLoc(), getThreePrimeLoc().getCoord()
        - getFivePrimeLoc().getCoord(), getBpReads());
  }

}
