/*
 * Sean Landman
 * Data Mining for Biomedical Informatics Group
 * University of Minnesota - Twin Cities
 */

package edu.umn.cs.kumarbio.shear.heterogeneity;

/**
 * BreakpointReads represents the respective number of reads that are clipped at
 * a particular breakpoint (in either direction) and that span the breakpoint.
 * This accounts for breakpoint ambiguity via microhomology by calculating the
 * clipped and spanning reads at the 5' and 3' edges of the possible breakpoint
 * locations.
 */
public class BreakpointReads {

  // The number of reads that are clipped in the 5' direction of this breakpoint
  // This means that they are clipped on the left side of the 5'-most edge of possible breakpoint locations
  private int fivePrimeClippedReads;

  // The number of reads that are clipped in the 3' direction of this breakpoint
  // This means that they are clipped on the right side of the 3'-most edge of possible breakpoint locations
  private int threePrimeClippedReads;

  // The number of reads that span the 5'-most edge of possible breakpoint locations
  private int fivePrimeSpanningReads;

  // The number of reads that span the 3'-most edge of possible breakpoint locations
  private int threePrimeSpanningReads;

  /**
   * Creates a new BreakpointReads object with the specified number of clipped
   * and spanning reads.
   *
   * @param fivePrimeClippedReads
   *          The number of reads that are clipped in the 5' direction of this
   *          breakpoint
   * @param fivePrimeClippedReads
   *          The number of reads that are clipped in the 3' direction of this
   *          breakpoint
   * @param fivePrimeSpanningReads
   *          The number of reads that span the 5'-most edge of possible
   *          breakpoint locations
   * @param threePrimeSpanningReads
   *          The number of reads that span the 3'-most edge of possible
   *          breakpoint locations
   */
  public BreakpointReads(
      int fivePrimeClippedReads, int threePrimeClippedReads, int fivePrimeSpanningReads, int threePrimeSpanningReads) {
    this.fivePrimeClippedReads = fivePrimeClippedReads;
    this.threePrimeClippedReads = threePrimeClippedReads;
    this.fivePrimeSpanningReads = fivePrimeSpanningReads;
    this.threePrimeSpanningReads = threePrimeSpanningReads;
  }

  /**
   * Returns the number of reads that are clipped in the 5' direction of this
   * breakpoint. This means that they are clipped on the left side of the
   * 5'-most edge of possible breakpoint locations.
   *
   * @return The number of reads that are clipped in the 5' direction of this
   *         breakpoint
   */
  public int getFivePrimeClippedReads() {
    return fivePrimeClippedReads;
  }

  /**
   * Returns the number of reads that are clipped in the 3' direction of this
   * breakpoint. This means that they are clipped on the right side of the
   * 3'-most edge of possible breakpoint locations
   *
   * @return The number of reads that are clipped in the 3' direction of this
   *         breakpoint
   */
  public int getThreePrimeClippedReads() {
    return threePrimeClippedReads;
  }

  /**
   * Returns the number of reads that span the 5'-most edge of possible
   * breakpoint locations.
   *
   * @return The number of reads that span the 5'-most edge of possible
   *         breakpoint locations
   */
  public int getFivePrimeSpanningReads() {
    return fivePrimeSpanningReads;
  }

  /**
   * Returns the number of reads that span the 3'-most edge of possible
   * breakpoint locations.
   *
   * @return The number of reads that span the 3'-most edge of possible
   *         breakpoint locations
   */
  public int getThreePrimeSpanningReads() {
    return threePrimeSpanningReads;
  }

  /**
   * Increments the number of reads that are clipped in the 5' direction of this
   * breakpoint by the specified amount.
   *
   * @param toAdd
   *          The amount to increment the 5' clipped reads by
   */
  public void addToFivePrimeClippedReads(int toAdd) {
    fivePrimeClippedReads += toAdd;
  }

  /**
   * Increments the number of reads that are clipped in the 3' direction of this
   * breakpoint by the specified amount.
   *
   * @param toAdd
   *          The amount to increment the 3' clipped reads by
   */
  public void addToThreePrimeClippedReads(int toAdd) {
    threePrimeClippedReads += toAdd;
  }

  /**
   * Increments the number of reads that span the 5'-most edge of possible
   * breakpoint locations.
   *
   * @param toAdd
   *          The amount to increment the 5' spanning reads by
   */
  public void addToFivePrimeSpanningReads(int toAdd) {
    fivePrimeSpanningReads += toAdd;
  }

  /**
   * Increments the number of reads that span the 3'-most edge of possible
   * breakpoint locations.
   *
   * @param toAdd
   *          The amount to increment the 3' spanning reads by
   */
  public void addToThreePrimeSpanningReads(int toAdd) {
    threePrimeSpanningReads += toAdd;
  }

  /**
   * Sets the number of reads that are clipped in the 5' direction of this
   * breakpoint to the specified amount.
   *
   * @param count
   *          The amount to set the 5' clipped reads to
   */
  public void setFivePrimeClippedReads(int count) {
    fivePrimeClippedReads = count;
  }

  /**
   * Sets the number of reads that are clipped in the 3' direction of this
   * breakpoint to the specified amount.
   *
   * @param count
   *          The amount to set the 3' clipped reads to
   */
  public void setThreePrimeClippedReads(int count) {
    threePrimeClippedReads = count;
  }

  /**
   * Sets the number of reads that span the 5'-most edge of possible breakpoint
   * locations to the specified amount.
   *
   * @param toAdd
   *          The amount to set the 5' spanning reads to
   */
  public void setFivePrimeSpanningReads(int count) {
    fivePrimeSpanningReads = count;
  }

  /**
   * Sets the number of reads that span the 3'-most edge of possible breakpoint
   * locations to the specified amount.
   *
   * @param toAdd
   *          The amount to set the 3' spanning reads to
   */
  public void setThreePrimeSpanningReads(int count) {
    threePrimeSpanningReads = count;
  }

  @Override
  public boolean equals(Object obj) {
    if (obj == null || !(obj instanceof BreakpointReads)) {
      return false;
    } else {
      BreakpointReads other = (BreakpointReads) obj;
      return getFivePrimeClippedReads() == other.getFivePrimeClippedReads()
          && getThreePrimeClippedReads() == other.getThreePrimeClippedReads()
          && getFivePrimeSpanningReads() == other.getFivePrimeSpanningReads()
          && getThreePrimeSpanningReads() == other.getThreePrimeSpanningReads();
    }
  }

  @Override
  public String toString() {
    return String.format(
        "%d/%d/%d/%d",
        getFivePrimeClippedReads(),
        getThreePrimeClippedReads(),
        getFivePrimeSpanningReads(),
        getThreePrimeSpanningReads());
  }

}
