/*
 * Sean Landman
 * Data Mining for Biomedical Informatics Group
 * University of Minnesota - Twin Cities
 */

package edu.umn.cs.kumarbio.shear.heterogeneity;

import java.io.File;
import java.io.FileNotFoundException;
import java.util.Scanner;
import java.util.Set;
import java.util.TreeSet;

import org.broadinstitute.gatk.utils.progressmeter.ProgressMeterDaemon;

import edu.umn.cs.kumarbio.shear.Shear;
import edu.umn.cs.kumarbio.shear.ShearRuntimeException;
import edu.umn.cs.kumarbio.shear.ShearSV;
import edu.umn.cs.kumarbio.shear.system.RunGatkCommand;
import edu.umn.cs.kumarbio.shear.variants.Deletion;
import edu.umn.cs.kumarbio.shear.variants.Insertion;
import edu.umn.cs.kumarbio.shear.variants.Inversion;
import edu.umn.cs.kumarbio.shear.variants.TandemDuplication;
import edu.umn.cs.kumarbio.shear.variants.Translocation;
import edu.umn.cs.kumarbio.shear.variants.Variant;

/**
 * SvHeterogeneityEstimator is used to estimate the heterogeneity for sets of
 * structural variants. It contains all of the logic for estimating
 * heterogeneity for different SV types.
 */
public class SvHeterogeneityEstimator {

  // The file path to the BAM alignment file to predict SV heterogeneity for
  private String  bamFilePath;

  // The file path to the FASTA reference file to predict SV heterogeneity for
  private String  faRefFilePath;

  // The instance of ShearSV that this heterogeneity estimator is called from
  private ShearSV shearSvInstance;

  /**
   * Creates a new SvHeterogeneityEstimator object.
   *
   * @param bamFilePath
   *          The file path to the BAM alignment file to predict SV
   *          heterogeneity for
   * @param faRefFilePath
   *          The file path to the FASTA reference file to predict SV
   *          heterogeneity for
   * @param shearSvInstance
   *          The instance of ShearSV that this heterogeneity estimator is
   *          called from
   */
  public SvHeterogeneityEstimator(String bamFilePath, String faRefFilePath, ShearSV shearSvInstance) {
    this.bamFilePath = bamFilePath;
    this.faRefFilePath = faRefFilePath;
    this.shearSvInstance = shearSvInstance;
  }

  /**
   * This estimates the heterogeneity for a set of structural variants.
   *
   * @param svs
   *          The set of structural variants to estimate the heterogeneity for
   */
  public void estimateHeterogeneity(TreeSet<Variant> svs) {

    // Loop through SVs
    for (Variant sv : svs) {

      // Set breakpoint reads information for the breakpoints of the SV
      if (sv instanceof Deletion || sv instanceof Insertion || sv instanceof TandemDuplication
          || sv instanceof Translocation) {
        for (Breakpoint bp : sv.getBreakpoints()) {
          setBreakpointReads(bp, false);
        }
      } else if (sv instanceof Inversion) {
        for (Breakpoint bp : sv.getBreakpoints()) {
          setBreakpointReads(bp, true);
        }
      } else {
        throw new ShearRuntimeException("Unexpected SV type encountered: " + sv);
      }

      // Calculate and set heterogeneity for SV and display
      sv.calculateHeterogeneity();
      StringBuilder svAndHetString = new StringBuilder();
      svAndHetString.append("  " + sv + " ");
      for (int i = 11 + sv.toString().length(); i < 93; i++) {
        svAndHetString.append(".");
      }
      svAndHetString.append(String.format("%6.2f%%", 100d * sv.getHeterogeneity()));
      Shear.log(svAndHetString.toString(), shearSvInstance.isDebugMode());

      // Get the number of still-running ProgressMeterDaemon threads
      Set<Thread> threads = Thread.getAllStackTraces().keySet();
      int numProgressMeterDaemonThreads = 0;
      for (Thread t : threads) {
        if (t instanceof ProgressMeterDaemon) {
          numProgressMeterDaemonThreads++;
        }
      }

      // Wait for ProgressMeterDaemon threads to finish if there are too many of them
      if (numProgressMeterDaemonThreads >= 10) {
        try {
          Thread.sleep(ProgressMeterDaemon.DEFAULT_POLL_FREQUENCY_MILLISECONDS);
        } catch (InterruptedException e) {
          throw new ShearRuntimeException(
              "Thread interrupted while waiting for GATK progress meter daemons to close.",
              e);
        }
      }

    }

  }

  /**
   * Runs the BreakpointReadsWalker GATK command for the given breakpoint and
   * sets its breakpoint reads information (i.e. the spanning and clipped reads
   * at that breakpoint).
   *
   * @param bp
   *          The breakpoint to run BreakpointReadsWalker on
   * @param unclippedSpanningOnly
   *          Whether or not to require spanning reads to be completely
   *          unclipped at all locations
   */
  private void setBreakpointReads(Breakpoint bp, boolean unclippedSpanningOnly) {

    // Temporary file to use
    String gatkOutputFilePath =
        shearSvInstance.getTmpFilePrefix() + "-" + bp.getFivePrimeLoc() + "-" + bp.getThreePrimeLoc() + ".txt";

    // Prepare BreakpointReadsWalker GATK arguments
    String gatkArgs =
        String.format(
            "-T BreakpointReadsWalker -I %s -R %s -fivePrimeEdgeBpLoc %d -threePrimeEdgeBpLoc %d -L %s:%d-%d -o %s",
            bamFilePath,
            faRefFilePath,
            bp.getFivePrimeLoc().getCoord(),
            bp.getThreePrimeLoc().getCoord(),
            bp.getFivePrimeLoc().getChr().getName(),
            bp.getFivePrimeLoc().getCoord(),
            bp.getThreePrimeLoc().getCoord() + 1,
            gatkOutputFilePath);
    if (unclippedSpanningOnly) {
      gatkArgs = gatkArgs + " -unclippedSpanningOnly";
    }
    if (shearSvInstance.isDebugMode()) {
      gatkArgs = gatkArgs + " -l INFO";
    } else {
      gatkArgs = gatkArgs + " -l ERROR";
    }

    // Run BreakpointReadsWalker GATK command
    RunGatkCommand.run(gatkArgs.split(" "), shearSvInstance.isDebugMode());

    // Number of clipped and spanning reads
    int fivePrimeClippedReads = 0;
    int threePrimeClippedReads = 0;
    int fivePrimeSpanningReads = 0;
    int threePrimeSpanningReads = 0;

    // Parse BreakpointReadsWalker output and delete output file
    Scanner sc = null;
    File gatkOutputFile = new File(gatkOutputFilePath);
    try {
      sc = new Scanner(gatkOutputFile);
      String outputFileContent = sc.nextLine();
      if (sc.hasNextLine() || !outputFileContent.matches("^[0-9]+\t[0-9]+\t[0-9]+\t[0-9]+$")) {
        throw new ShearRuntimeException("Unexpected format in BreakpointReadsWalker output file: " + gatkOutputFilePath);
      }
      String[] tokens = outputFileContent.split("\t");
      fivePrimeClippedReads = Integer.valueOf(tokens[0]);
      threePrimeClippedReads = Integer.valueOf(tokens[1]);
      fivePrimeSpanningReads = Integer.valueOf(tokens[2]);
      threePrimeSpanningReads = Integer.valueOf(tokens[3]);
    } catch (FileNotFoundException e) {
      throw new ShearRuntimeException("Problem reading from the BreakpointReadsWalker output file: "
          + gatkOutputFilePath, e);
    } finally {
      sc.close();
      gatkOutputFile.delete();
    }

    // Set new breakpoint reads information for this breakpoint
    bp.getBpReads().setFivePrimeClippedReads(fivePrimeClippedReads);
    bp.getBpReads().setThreePrimeClippedReads(threePrimeClippedReads);
    bp.getBpReads().setFivePrimeSpanningReads(fivePrimeSpanningReads);
    bp.getBpReads().setThreePrimeSpanningReads(threePrimeSpanningReads);

  }

}
