/*
 * Sean Landman
 * Data Mining for Biomedical Informatics Group
 * University of Minnesota - Twin Cities
 */

package edu.umn.cs.kumarbio.shear.realignment;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;

import edu.umn.cs.kumarbio.Genome;
import edu.umn.cs.kumarbio.shear.ShearRuntimeException;
import edu.umn.cs.kumarbio.shear.variants.Location;

/**
 * RealignmentRegionsSet keeps track of the set of all realignment regions for a
 * particular iteration of local breakpoint read extraction and realignment.
 * Realignment regions must have a certain amount of bases in between them,
 * which is specified by the regionSeparation variable.
 */
public class RealignmentRegionsSet {

  // The Genome object to use
  private Genome                                        gen;

  // Data structure to keep track of this set of realignment regions
  // HashMap key is the chromosome name, and the array is an ordered list of realignment regions
  private HashMap<String, ArrayList<RealignmentRegion>> regions;

  // The number of bases that must be in between two adjacent realignment regions (otherwise they will be merged)
  private long                                          regionSeparation;

  /**
   * Creates a new RealignmentRegionsSet object by initializing an empty set of
   * realignment regions and setting the Genome object and the specified
   * regionSeparation variable.
   *
   * @param gen
   *          The Genome object to use
   * @param regionSeparation
   *          The number of bases that must be in between two adjacent
   *          realignment regions (otherwise they will be merged)
   */
  public RealignmentRegionsSet(Genome gen, long regionSeparation) {
    this.gen = gen;
    regions = new HashMap<String, ArrayList<RealignmentRegion>>();
    this.regionSeparation = regionSeparation;
  }

  /**
   * Returns a list of all realignment regions in this set.
   *
   * @return An ArrayList containing all of the realignment regions in this set.
   */
  public ArrayList<RealignmentRegion> getAllRealignmentRegions() {
    ArrayList<RealignmentRegion> allRegions = new ArrayList<RealignmentRegion>();
    for (Map.Entry<String, ArrayList<RealignmentRegion>> entry : regions.entrySet()) {
      for (RealignmentRegion currentRegion : entry.getValue()) {
        allRegions.add(currentRegion);
      }
    }
    return allRegions;
  }

  /**
   * Returns the realignment region for the specified read. Reads are specified
   * using their alignment chromosome and start location.
   *
   * @param chrName
   *          The name of the chromosome that the specified read is aligned to
   * @param pos
   *          The alignment coordinate for the specified read
   * @return The realignment region for the specified read
   */
  public RealignmentRegion getRealignmentRegionForRead(String chrName, long pos) {

    // Realignment region to return
    RealignmentRegion regionToReturn = null;

    // Throw error for unexpected chromosome name
    if (!regions.containsKey(chrName)) {
      throw new ShearRuntimeException("Chromosome '" + chrName
          + "' is expected to be in realignment region set, but is not");
    }

    // Whether or not the appropriate realignment region was found
    boolean regionFound = false;

    // Find the appropriate realignment region for this read
    for (RealignmentRegion currentRegion : regions.get(chrName)) {
      if (currentRegion.getStartLoc().getCoord() <= pos && currentRegion.getEndLoc().getCoord() >= pos) {
        if (!regionFound) {
          regionFound = true;
          regionToReturn = currentRegion;
        } else {
          throw new ShearRuntimeException("Read aligned at '" + chrName + ":" + pos
              + "' matches multiple realignment regions");
        }
      }
    }

    // Throw error if appropriate realignment region was not found
    if (!regionFound) {
      throw new ShearRuntimeException("Read aligned at '" + chrName + ":" + pos
          + "' is expected to match a realignment region, but does not");
    }

    // Return realignment region
    return regionToReturn;

  }

  /**
   * Updates the current set of realignment regions by processing a new read.
   * The alignment chromosome and position, as well as the read's CIGAR string
   * are processed to determine if a new realignment region needs to be added or
   * a current realignment region needs to be expanded (possibly merging
   * realignment regions that become too close to one another).
   *
   * @param chrName
   *          The name of the chromosome that the read to process is aligned to
   * @param pos
   *          The alignment coordinate for the read to process
   * @param cigar
   *          The CIGAR string for the read to process
   */
  public void addRead(String chrName, long pos, String cigar) {

    // If new chromosome, add a new array of regions
    if (!regions.containsKey(chrName)) {
      regions.put(chrName, new ArrayList<RealignmentRegion>());
    }

    // Validate that CIGAR is valid
    if (!cigar.matches("([0-9]+(M|I|D|N|S|H|P|=|X))+")) {
      throw new ShearRuntimeException("Invalid CIGAR string encountered: " + cigar);
    }

    // Initialize start and end position for this read with soft-clipping included (determine exact coordinates below)
    long startPos = pos;
    long endPos = pos;

    // Whether or not we are processing the first non-H CIGAR element
    boolean firstCigarEl = true;

    // String to store numbers for lenth of current CIGAR element
    String currentNumStr = "";

    // Loop through CIGAR string
    for (int i = 0; i < cigar.length(); i++) {
      char c = cigar.charAt(i);

      // Update string of current CIGAR element length
      if (c >= '0' && c <= '9') {
        currentNumStr += c;

        // If current character is an element symbol, parse and reset length string and process
      } else {
        long currentNum = Long.valueOf(currentNumStr);
        currentNumStr = "";

        // Ignore H elements and don't update firstCigarEl
        if (c == 'H') {
          continue;
        }

        // If S is first, then update startPos by subtracting soft-clip length
        // If S is not first, then update endPos by adding soft-clip length
        // For M/D/N/=/X, add the length to endPos
        // Ignore I/P but do update firstCigarEl
        switch (c) {
          case 'S':
            if (firstCigarEl) {
              startPos -= currentNum;
            } else {
              endPos += currentNum;
            }
            break;
          case 'M':
          case 'D':
          case 'N':
          case '=':
          case 'X':
            endPos += currentNum;
          case 'I':
          case 'P':
            break;
          default:
            throw new ShearRuntimeException("Invalid CIGAR string encountered: " + cigar);
        }
        firstCigarEl = false;

        // Fix endPos by subtracting 1
        endPos -= 1;

      }

    }

    // If no realignment regions for this chromosome yet, add one
    if (regions.get(chrName).size() == 0) {
      regions.get(chrName).add(
          new RealignmentRegion(new Location(gen.getChromosome(chrName), startPos), new Location(gen
              .getChromosome(chrName), endPos)));

      // Otherwise, figure out how to update current set of realignment regions
    } else {

      // Whether or not the current read region was handled
      boolean readHandled = false;

      // Loop through current set of realignment regions
      for (int i = 0; i < regions.get(chrName).size(); i++) {
        long currentRegionStart = regions.get(chrName).get(i).getStartLoc().getCoord();
        long currentRegionEnd = regions.get(chrName).get(i).getEndLoc().getCoord();

        // If read region is before the current realignment region, add a new realignment region
        // We already know it's clear of all previous realignment regions
        if (endPos < currentRegionStart - regionSeparation) {
          regions.get(chrName).add(
              i,
              new RealignmentRegion(new Location(gen.getChromosome(chrName), startPos), new Location(gen
                  .getChromosome(chrName), endPos)));
          readHandled = true;
          break;

          // If read region is within regionSeparation of current realignment region, then:
          // 1 - Possibly update the current realignment region coordinates
          // 2 - Possibly merge current realignment region with one or both of its neighbors
        } else if (endPos >= currentRegionStart - regionSeparation && startPos <= currentRegionEnd + regionSeparation) {
          regions.get(chrName).get(i).updateStartCoord(Math.min(startPos, currentRegionStart));
          regions.get(chrName).get(i).updateEndCoord(Math.max(endPos, currentRegionEnd));
          if (i < regions.get(chrName).size() - 1
              && regions.get(chrName).get(i + 1).getStartLoc().getCoord()
                  - regions.get(chrName).get(i).getEndLoc().getCoord() <= regionSeparation) {
            regions.get(chrName).get(i).updateEndCoord(regions.get(chrName).get(i + 1).getEndLoc().getCoord());
            regions.get(chrName).remove(i + 1);
          }
          if (i > 0
              && regions.get(chrName).get(i).getStartLoc().getCoord()
                  - regions.get(chrName).get(i - 1).getEndLoc().getCoord() <= regionSeparation) {
            regions.get(chrName).get(i).updateStartCoord(regions.get(chrName).get(i - 1).getStartLoc().getCoord());
            regions.get(chrName).remove(i - 1);
          }
          readHandled = true;
          break;

        }

      }

      // If read region was not handled already, add a new realignment region to the end of the list
      if (!readHandled) {
        regions.get(chrName).add(
            new RealignmentRegion(new Location(gen.getChromosome(chrName), startPos), new Location(gen
                .getChromosome(chrName), endPos)));
      }

    }

  }

}
