/*
 * Sean Landman
 * Data Mining for Biomedical Informatics Group
 * University of Minnesota - Twin Cities
 */

package edu.umn.cs.kumarbio.shear.variantcallers;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Scanner;
import java.util.TreeSet;

import edu.umn.cs.kumarbio.Genome;
import edu.umn.cs.kumarbio.shear.Shear;
import edu.umn.cs.kumarbio.shear.ShearRuntimeException;
import edu.umn.cs.kumarbio.shear.system.RunSystemCommand;
import edu.umn.cs.kumarbio.shear.variants.Deletion;
import edu.umn.cs.kumarbio.shear.variants.IllegalVariantException;
import edu.umn.cs.kumarbio.shear.variants.Inversion;
import edu.umn.cs.kumarbio.shear.variants.Location;
import edu.umn.cs.kumarbio.shear.variants.TandemDuplication;
import edu.umn.cs.kumarbio.shear.variants.Translocation;
import edu.umn.cs.kumarbio.shear.variants.Variant;

/**
 * This is a wrapper for the DELLY SV caller. This class extends the abstract
 * class VariantCaller.
 */
public class DellyWrapper extends VariantCaller {

  // The Genome object to use for this instance of DELLY
  private Genome  gen;

  // Whether or not debugging mode is active
  private boolean debug;

  public DellyWrapper(Genome gen, boolean debug) {

    // Set instance variables
    this.gen = gen;
    this.debug = debug;

  }

  @Override
  public String runVariantCaller(String bamFile, String outputPrefix, List<String> options) {

    // Combined output file to use
    String combinedDellyOutputFilePath = outputPrefix + ".COMBINED.vcf";

    // List of DELLY SV types / prediction modes
    List<String> dellyModes = Arrays.asList("DEL", "DUP", "INV", "TRA");

    // Delete files that may be present from a previous run
    new File(combinedDellyOutputFilePath).delete();
    for (String dellyMode : dellyModes) {
      new File(outputPrefix + "." + dellyMode + ".vcf").delete();
    }

    // Run DELLY for each DELLY SV type / prediction mode
    for (String dellyMode : dellyModes) {
      ArrayList<String> dellyCommandAndArgs = new ArrayList<String>();
      dellyCommandAndArgs.add("delly");
      dellyCommandAndArgs.add("-t");
      dellyCommandAndArgs.add(dellyMode);
      dellyCommandAndArgs.add("-o");
      dellyCommandAndArgs.add(outputPrefix + "." + dellyMode + ".vcf");
      dellyCommandAndArgs.add("-g");
      dellyCommandAndArgs.add(gen.getFaFilePath());
      for (int i = 0; i < options.size(); i++) {
        dellyCommandAndArgs.add(options.get(i));
      }
      dellyCommandAndArgs.add(bamFile);
      ProcessBuilder dellyProcess = new ProcessBuilder(dellyCommandAndArgs);
      RunSystemCommand.run(dellyProcess, System.out, "DELLY", debug);
    }

    try {

      // Whether or not we have copied the DELLY VCF headers
      boolean copiedHeaders = false;

      // Output stream to combined output file
      PrintWriter combinedOutput = new PrintWriter(combinedDellyOutputFilePath);

      // Loop through existing DELLY result files
      for (String dellyMode : dellyModes) {
        String dellyResultFile = outputPrefix + "." + dellyMode + ".vcf";
        if (new File(dellyResultFile).exists()) {

          // Loop through this DELLY result file, adding the headers if it is the first, and adding variant calls regardless
          Scanner sc = new Scanner(new File(dellyResultFile));
          while (sc.hasNextLine()) {
            String line = sc.nextLine();
            if (line.startsWith("#") && !copiedHeaders) {
              combinedOutput.println(line);
            } else if (!line.startsWith("#")) {
              copiedHeaders = false;
              combinedOutput.println(line);
            }
          }
          sc.close();

        }
      }

      // Close output stream to combined output file
      combinedOutput.close();

      // Catch errors
    } catch (FileNotFoundException e) {
      throw new ShearRuntimeException("Error creating combined DELLY output file", e);
    }

    // Clean up extra output files if not in debug mode
    if (!debug) {
      for (String dellyMode : dellyModes) {
        new File(outputPrefix + "." + dellyMode + ".vcf").delete();
      }
    }

    // Return combined output file path
    return combinedDellyOutputFilePath;

  }

  @Override
  public TreeSet<Variant> getVariants(String variantCallerOutputFile) {

    // Note that this getVariants implementation DOES NOT calculate heterogeneity percentages for variants

    // Initialize set of variants to return
    TreeSet<Variant> variants = new TreeSet<Variant>();

    // Open up a scanner for VCF output file and read from it
    Scanner sc = null;
    try {
      sc = new Scanner(new File(variantCallerOutputFile));
      while (sc.hasNextLine()) {
        String line = sc.nextLine();
        if (!line.startsWith("#")) {

          // Validate that line has expected format
          String[] tokens1 = line.split("\t");
          if (tokens1.length >= 8 && tokens1[1].matches("^[0-9]+$") && tokens1[4].matches("^<DEL>|<DUP>|<INV>|<TRA>|$")) {
            String[] tokens2 = tokens1[7].split(";");
            if (tokens2.length >= 8 && tokens2[0].matches("^PRECISE|IMPRECISE$")
                && tokens2[3].matches("^SVTYPE=(DEL|DUP|INV|TRA)$") && tokens2[5].startsWith("CHR2=")
                && tokens2[6].matches("^END=[0-9]+$") && tokens2[7].matches("^CT=(3to5|5to3|3to3|5to5)$")) {

              // Get relevant VCF line information
              String chr1 = tokens1[0];
              long loc1 = Long.valueOf(tokens1[1]);
              String svType1 = tokens1[4].substring(1, 4);
              String precision = tokens2[0];
              String svType2 = tokens2[3].substring(7);
              String chr2 = tokens2[5].substring(5);
              long loc2 = Long.valueOf(tokens2[6].substring(4));
              String ct = tokens2[7].substring(3);

              // The variant object
              Variant v = null;

              try {

                // Create Deletion
                if (svType1.equals("DEL") && svType2.equals("DEL") && precision.equals("PRECISE") && chr1.equals(chr2)
                    && loc1 < loc2 && ct.equals("3to5")) {
                  v =
                      new Deletion(new Location(gen.getChromosome(chr1), loc1 + 1), loc2 - loc1, false)
                          .getCanonicalEquivalent();

                  // Create Tandem Duplication
                } else if (svType1.equals("DUP") && svType2.equals("DUP") && precision.equals("PRECISE")
                    && chr1.equals(chr2) && loc1 < loc2 && ct.equals("5to3")) {
                  v =
                      new TandemDuplication(new Location(gen.getChromosome(chr1), loc1 + 1), loc2 - loc1)
                          .getCanonicalEquivalent();

                  // Create Inversion
                } else if (svType1.equals("INV") && svType2.equals("INV") && precision.equals("PRECISE")
                    && chr1.equals(chr2) && loc1 < loc2 && (ct.equals("5to5") || ct.equals("3to3"))) {
                  v =
                      new Inversion(new Location(gen.getChromosome(chr1), loc1 + 1), loc2 - loc1)
                          .getCanonicalEquivalent();

                  // Create Translocation (type 1,  3to5)
                } else if (svType1.equals("TRA") && svType2.equals("TRA") && precision.equals("PRECISE")
                    && ct.equals("3to5")) {
                  v =
                      new Translocation(new Location(gen.getChromosome(chr1), loc1), new Location(
                          gen.getChromosome(chr2),
                          loc2 + 1), Translocation.Type.for3toFor5).getCanonicalEquivalent();

                  // Create Translocation (type 2,  5to3)
                } else if (svType1.equals("TRA") && svType2.equals("TRA") && precision.equals("PRECISE")
                    && ct.equals("5to3")) {
                  v =
                      new Translocation(new Location(gen.getChromosome(chr2), loc2), new Location(
                          gen.getChromosome(chr1),
                          loc1 + 1), Translocation.Type.for3toFor5).getCanonicalEquivalent();

                  // Create Translocation (type 3,  3to3)
                } else if (svType1.equals("TRA") && svType2.equals("TRA") && precision.equals("PRECISE")
                    && ct.equals("3to3")) {
                  v =
                      new Translocation(new Location(gen.getChromosome(chr2), loc2), new Location(
                          gen.getChromosome(chr1),
                          loc1), Translocation.Type.for3toRev5).getCanonicalEquivalent();

                  // Create Translocation (type 3,  5to5)
                } else if (svType1.equals("TRA") && svType2.equals("TRA") && precision.equals("PRECISE")
                    && ct.equals("5to5")) {
                  v =
                      new Translocation(new Location(gen.getChromosome(chr2), loc2 + 1), new Location(
                          gen.getChromosome(chr1),
                          loc1 + 1), Translocation.Type.for5toRev3).getCanonicalEquivalent();

                  // Skip over lines with IMPRECISE tag (but no warning message)
                } else if (precision.equals("IMPRECISE")) {
                  continue;

                  // Skip over lines with unexpected format
                } else {
                  Shear.log("WARNING: Skipping unexpected event in DELLY output (" + tokens1[0] + "\t" + tokens1[1]
                      + "\t" + tokens1[4] + "\t" + tokens2[0] + "\t" + tokens2[3] + "\t" + tokens2[5] + "\t"
                      + tokens2[6] + "\t" + tokens2[7] + ")", debug);
                  continue;

                }

                // Skip over line that caused IllegalVariantException error
              } catch (IllegalVariantException e) {
                Shear.log("WARNING: Skipping illegal variant in DELLY output (" + tokens1[0] + "\t" + tokens1[1] + "\t"
                    + tokens1[4] + "\t" + tokens2[0] + "\t" + tokens2[3] + "\t" + tokens2[5] + "\t" + tokens2[6] + "\t"
                    + tokens2[7] + ")", debug);
                continue;

              }

              // Add variant to return set
              variants.add(v);

              // Throw errors for unexpected VCF format
            } else {
              throw new ShearRuntimeException("Unexpected format in VCF output file '" + variantCallerOutputFile
                  + "' from DELLY");
            }
          } else {
            throw new ShearRuntimeException("Unexpected format in VCF output file '" + variantCallerOutputFile
                + "' from DELLY");
          }

        }
      }

      // Catch error from setting up Scanner object for VCF output
    } catch (FileNotFoundException e) {
      throw new ShearRuntimeException(
          "Unable to locate VCF output file '" + variantCallerOutputFile + "' from DELLY",
          e);

    } finally {

      // Close Scanner object
      if (sc != null) {
        sc.close();
      }

      // Clean up VCF output file if not in debug mode
      if (!debug) {
        new File(variantCallerOutputFile).delete();
      }

    }

    // Return variants
    return variants;

  }

}
