/*
 * Sean Landman
 * Data Mining for Biomedical Informatics Group
 * University of Minnesota - Twin Cities
 */

package edu.umn.cs.kumarbio.shear.variantcallers;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.Scanner;
import java.util.TreeSet;

import edu.umn.cs.kumarbio.Genome;
import edu.umn.cs.kumarbio.shear.Shear;
import edu.umn.cs.kumarbio.shear.ShearRuntimeException;
import edu.umn.cs.kumarbio.shear.system.RunGatkCommand;
import edu.umn.cs.kumarbio.shear.variants.Deletion;
import edu.umn.cs.kumarbio.shear.variants.IllegalVariantException;
import edu.umn.cs.kumarbio.shear.variants.Insertion;
import edu.umn.cs.kumarbio.shear.variants.Location;
import edu.umn.cs.kumarbio.shear.variants.SNP;
import edu.umn.cs.kumarbio.shear.variants.Variant;

/**
 * This is a wrapper for the GATK HaplotypeCaller. This class extends the
 * abstract class VariantCaller.
 */
public class HaplotypeCallerWrapper extends VariantCaller {

  // The Genome object to use for this haplotype caller
  private Genome  gen;

  // Whether or not debugging mode is active
  private boolean debug;

  /**
   * Creates a new HaplotypeCallerWrapper using the specified genome. The FASTA
   * reference file associated the genome is required to have an accompanying
   * *.fai index file.
   *
   * @param gen
   *          The Genome object to use for this haplotype caller
   * @param debug
   *          Whether or not debugging mode is active
   */
  public HaplotypeCallerWrapper(Genome gen, boolean debug) {

    // Check for existence of FASTA file and its index
    if (!new File(gen.getFaFilePath()).exists()) {
      throw new ShearRuntimeException("Required file does not exist: " + gen.getFaFilePath());
    } else if (!new File(gen.getFaFilePath() + ".fai").exists()) {
      throw new ShearRuntimeException("Required file does not exist: " + gen.getFaFilePath() + ".fai");
    }

    // Set instance variables
    this.gen = gen;
    this.debug = debug;

  }

  @Override
  public String runVariantCaller(String bamFile, String outputPrefix, List<String> options) {

    // Output file to use
    String outputFilePath = outputPrefix + ".vcf";

    // Set up GATK arguments array
    String[] gatkArgs = new String[12 + options.size()];
    int ind = 0;
    gatkArgs[ind++] = "-T";
    gatkArgs[ind++] = "HaplotypeCaller";
    gatkArgs[ind++] = "-I";
    gatkArgs[ind++] = bamFile;
    gatkArgs[ind++] = "-R";
    gatkArgs[ind++] = gen.getFaFilePath();
    gatkArgs[ind++] = "-o";
    gatkArgs[ind++] = outputFilePath;
    gatkArgs[ind++] = "-rf";
    gatkArgs[ind++] = "BadCigar";
    gatkArgs[ind++] = "-l";
    if (debug) {
      gatkArgs[ind++] = "INFO";
    } else {
      gatkArgs[ind++] = "ERROR";
    }
    for (int i = 0; i < options.size(); i++) {
      gatkArgs[ind++] = options.get(i);
    }

    // Run GATK HaplotypeCaller
    RunGatkCommand.run(gatkArgs, debug);

    // Return output file path
    return outputFilePath;

  }

  @Override
  public TreeSet<Variant> getVariants(String variantCallerOutputFile) {

    // Note that this getVariants implementation DOES calculate heterogeneity percentages for variants

    // Initialize set of variants to return
    TreeSet<Variant> variants = new TreeSet<Variant>();

    // Open up a scanner for VCF output file and read from it
    Scanner sc = null;
    try {
      sc = new Scanner(new File(variantCallerOutputFile));
      newVcfLine: while (sc.hasNextLine()) {
        String line = sc.nextLine();

        // Get non-header line
        if (line.charAt(0) != '#') {
          String[] tokens = line.split("\t");

          // Validate that line has expected format
          if (tokens.length >= 10 && tokens[1].matches("^[0-9]+$")) {

            // Get relevant VCF line information
            String chr = tokens[0];
            long loc = Long.valueOf(tokens[1]);
            String refBases = tokens[3];
            String varBases = tokens[4];
            String[] formatTags = tokens[8].split(":");
            String[] formatInfo = tokens[9].split(":");

            // The variant object(s) from this VCF line
            ArrayList<Variant> currentVcfLineVariants = new ArrayList<Variant>();

            // The list of alleles associated with each variant from this VCF line (in case multiple SNPs in same varBases)
            ArrayList<Integer> alleles = new ArrayList<Integer>();

            try {

              // Validate that refBases contains only ACGT and varBases contains only comma-separated ACGT
              if (refBases.matches("^[ACGT]+$") && varBases.matches("^[ACGT]+(,[ACGT]+)*$")
                  && gen.getSequence(chr, loc, loc + refBases.length() - 1).equals(refBases)) {

                // Split varBases
                String[] allVarBases = varBases.split(",");

                // Make a different Variant object for each entry in varBases
                for (int allele = 1; allele <= allVarBases.length; allele++) {
                  String currentVarBases = allVarBases[allele - 1];
                  int variantSize = currentVarBases.length() - refBases.length();

                  // Handle SNP(s)
                  if (variantSize == 0) {

                    // Loop through bases, adding a SNP object for each differing base between ref and var bases
                    for (int i = 0; i < currentVarBases.length(); i++) {
                      if (currentVarBases.charAt(i) != refBases.charAt(i)) {
                        currentVcfLineVariants.add(new SNP(
                            new Location(gen.getChromosome(chr), loc + i),
                            currentVarBases.charAt(i)).getCanonicalEquivalent());
                        alleles.add(allele);
                      }
                    }

                    // Handle Insertion
                  } else if (variantSize > 0 && refBases.charAt(0) == currentVarBases.charAt(0)) {

                    // Find number of identical start bases between ref and var bases
                    int identicalStartBases = 1;
                    while (identicalStartBases + 1 <= refBases.length()
                        && identicalStartBases + 1 <= currentVarBases.length()
                        && refBases.substring(0, identicalStartBases + 1).equals(
                            currentVarBases.substring(0, identicalStartBases + 1))) {
                      identicalStartBases++;
                    }

                    // Find number of identical end bases between ref and var bases
                    int identicalEndBases = 0;
                    while (identicalEndBases + 1 <= refBases.length() - 1
                        && identicalEndBases + 1 <= currentVarBases.length() - 1
                        && refBases.substring(refBases.length() - identicalEndBases - 1, refBases.length()).equals(
                            currentVarBases.substring(
                                currentVarBases.length() - identicalEndBases - 1,
                                currentVarBases.length()))) {
                      identicalEndBases++;
                    }

                    // Ensure that identical bases add up to at least the length of ref bases
                    if (identicalStartBases + identicalEndBases < refBases.length()) {
                      Shear.log("WARNING: Skipping unexpected event in HaplotypeCaller output (" + chr + ":" + loc
                          + " " + refBases + " -> " + varBases + "), insertion is invalid", debug);
                      continue newVcfLine;
                    }

                    // Get coordinate for insertion and insertion string
                    int preInsertionBasesInVarBases = identicalStartBases;
                    int postInsertionBasesInVarBases = identicalEndBases;
                    while (preInsertionBasesInVarBases + postInsertionBasesInVarBases != refBases.length()) {
                      if (postInsertionBasesInVarBases > 0) {
                        postInsertionBasesInVarBases--;
                      } else {
                        preInsertionBasesInVarBases--;
                      }
                    }
                    long insertionLoc = loc + preInsertionBasesInVarBases;
                    String insertionString =
                        currentVarBases.substring(preInsertionBasesInVarBases, currentVarBases.length()
                            - postInsertionBasesInVarBases);

                    // Add Insertion object
                    currentVcfLineVariants.add(new Insertion(
                        new Location(gen.getChromosome(chr), insertionLoc),
                        insertionString,
                        true).getCanonicalEquivalent());
                    alleles.add(allele);

                    // Handle Deletion
                  } else if (variantSize < 0 && refBases.charAt(0) == currentVarBases.charAt(0)) {

                    // Find number of identical start bases between ref and var bases
                    int identicalStartBases = 1;
                    while (identicalStartBases + 1 <= refBases.length()
                        && identicalStartBases + 1 <= currentVarBases.length()
                        && refBases.substring(0, identicalStartBases + 1).equals(
                            currentVarBases.substring(0, identicalStartBases + 1))) {
                      identicalStartBases++;
                    }

                    // Find number of identical end bases between ref and var bases
                    int identicalEndBases = 0;
                    while (identicalEndBases + 1 <= refBases.length() - 1
                        && identicalEndBases + 1 <= currentVarBases.length() - 1
                        && refBases.substring(refBases.length() - identicalEndBases - 1, refBases.length()).equals(
                            currentVarBases.substring(
                                currentVarBases.length() - identicalEndBases - 1,
                                currentVarBases.length()))) {
                      identicalEndBases++;
                    }

                    // Ensure that identical bases add up to at least the length of deleted bases
                    if (identicalStartBases + identicalEndBases < currentVarBases.length()) {
                      Shear.log("WARNING: Skipping unexpected event in HaplotypeCaller output (" + chr + ":" + loc
                          + " " + refBases + " -> " + varBases + "), deletion is invalid", debug);
                      continue newVcfLine;
                    }

                    // Get coordinate for deletion
                    int preDeletionBasesInVarBases = identicalStartBases;
                    int postDeletionBasesInVarBases = identicalEndBases;
                    while (preDeletionBasesInVarBases + postDeletionBasesInVarBases != currentVarBases.length()) {
                      if (postDeletionBasesInVarBases > 0) {
                        postDeletionBasesInVarBases--;
                      } else {
                        preDeletionBasesInVarBases--;
                      }
                    }
                    long deletionLoc = loc + preDeletionBasesInVarBases;

                    // Add Deletion object
                    currentVcfLineVariants.add(new Deletion(
                        new Location(gen.getChromosome(chr), deletionLoc),
                        -variantSize,
                        true).getCanonicalEquivalent());
                    alleles.add(allele);

                    // Skip over lines with unexpected format
                  } else {
                    Shear.log("WARNING: Skipping unexpected event in HaplotypeCaller output (" + chr + ":" + loc + " "
                        + refBases + " -> " + varBases + "), unable to parse into a variant object", debug);
                    continue newVcfLine;
                  }

                }

                // Skip over lines with unexpected format
              } else {
                Shear.log("WARNING: Skipping unexpected event in HaplotypeCaller output (" + chr + ":" + loc + " "
                    + refBases + " -> " + varBases + "), reference and/or variant bases were not correct", debug);
                continue newVcfLine;
              }

              // Skip over line that caused IllegalVariantException error
            } catch (IllegalVariantException e) {
              Shear.log("[DEBUG] WARNING: Skipping illegal variant in HaplotypeCaller output (" + chr + ":" + loc + " "
                  + refBases + " -> " + varBases + ")", debug);
              continue newVcfLine;

              // Throw error for Genome I/O problems
            } catch (IOException e) {
              throw new ShearRuntimeException("Unable to read from FASTA reference file", e);
            }

            // Calculate heterogeneity for all variants for this VCF line
            for (int i = 0; i < formatTags.length; i++) {
              if (formatTags[i].equals("AD") && formatInfo[i].matches("^[0-9]+,[0-9]+(,[0-9]+)*$")) {
                String[] alleleDepths = formatInfo[i].split(",");
                if (alleleDepths.length != alleles.size() + 1) {
                  Shear.log("WARNING: Skipping unexpected event in HaplotypeCaller output (" + chr + ":" + loc + " "
                      + refBases + " -> " + varBases + "), AD tag is missing or malformed", debug);
                  continue newVcfLine;
                }
                double totalDepth = 0d;
                for (int j = 0; j < alleleDepths.length; j++) {
                  totalDepth += Double.valueOf(alleleDepths[j]);
                }
                for (int j = 0; j < currentVcfLineVariants.size(); j++) {
                  double varDepth = Double.valueOf(alleleDepths[alleles.get(j)]);
                  double heterogeneity = varDepth / totalDepth;
                  if (heterogeneity > 1d) {
                    heterogeneity = 1d;
                  }
                  currentVcfLineVariants.get(j).setHeterogeneity(heterogeneity);
                }
                break;
              }
            }

            // Display warning if AD tag not found or malformed (i.e. heterogeneity is still -1 for any of the variants)
            for (Variant v : currentVcfLineVariants) {
              if (v.getHeterogeneity() < 0d) {
                Shear.log("WARNING: Skipping unexpected event in HaplotypeCaller output (" + chr + ":" + loc + " "
                    + refBases + " -> " + varBases + "), AD tag is missing or malformed", debug);
                continue newVcfLine;
              }
            }

            // Add variants to return set if they have more than 0% heterogeneity
            for (Variant v : currentVcfLineVariants) {
              if (v.getHeterogeneity() > 0d) {
                variants.add(v);
              } else if (debug) {
                Shear.log("Filtering out SNP/INDEL (" + chr + ":" + loc + " " + refBases + " -> " + varBases
                    + ") with 0% heterogeneity, considered false positive", debug);
              }
            }

            // Throw error for unexpected VCF format
          } else {
            throw new ShearRuntimeException("Unexpected format in VCF output file '" + variantCallerOutputFile
                + "' from HaplotypeCaller");
          }

        }

      }

      // Catch error from setting up Scanner object for VCF output
    } catch (FileNotFoundException e) {
      throw new ShearRuntimeException("Unable to locate VCF output file '" + variantCallerOutputFile
          + "' from HaplotypeCaller", e);

    } finally {

      // Close Scanner object
      if (sc != null) {
        sc.close();
      }

      // Clean up VCF output files if not in debug mode
      if (!debug) {
        new File(variantCallerOutputFile).delete();
        new File(variantCallerOutputFile + ".idx").delete();
      }

    }

    // Return variants
    return variants;

  }

}
