/*
 * Sean Landman
 * Data Mining for Biomedical Informatics Group
 * University of Minnesota - Twin Cities
 */

package edu.umn.cs.kumarbio.shear.variantcallers;

import java.util.ArrayList;
import java.util.List;
import java.util.TreeSet;

import edu.umn.cs.kumarbio.shear.variants.Variant;

/**
 * Represents a program that is used to call variants.
 */
public abstract class VariantCaller {

  /**
   * Runs this variant caller on the specified BAM file.
   *
   * @param bamFile
   *          The location of the BAM file to call variants on
   * @param outputPrefix
   *          The prefix to use for any output files that may be produced
   * @return The file path to the final output file for this instance of a
   *         variant caller
   */
  public String runVariantCaller(String bamFile, String outputPrefix) {
    return runVariantCaller(bamFile, outputPrefix, new ArrayList<String>());
  }

  /**
   * Runs this variant caller on the specified BAM file with the specified
   * options.
   *
   * @param bamFile
   *          The location of the BAM file to call variants on
   * @param outputPrefix
   *          The prefix to use for any output files that may be produced
   * @param options
   *          The additional options to use while running this variant caller
   * @return The file path to the final output file for this instance of a
   *         variant caller
   */
  public abstract String runVariantCaller(String bamFile, String outputPrefix, List<String> options);

  /**
   * Returns the ordered set of variants from the specified output file produced
   * by this variant caller.
   *
   * @param variantCallerOutputFile
   *          The variant caller output file to parse
   * @return The ordered set of variants called from the specified variant
   *         caller output file
   */
  public abstract TreeSet<Variant> getVariants(String variantCallerOutputFile);

  /**
   * Closes this instance of a variant caller. The default close method defined
   * here does nothing.
   */
  public void close() {
  }

}
