/*
 * Sean Landman
 * Data Mining for Biomedical Informatics Group
 * University of Minnesota - Twin Cities
 */

package edu.umn.cs.kumarbio.shear.variants;

import java.io.IOException;

import edu.umn.cs.kumarbio.Genome;
import edu.umn.cs.kumarbio.shear.ShearRuntimeException;
import edu.umn.cs.kumarbio.shear.heterogeneity.Breakpoint;
import edu.umn.cs.kumarbio.shear.heterogeneity.BreakpointReads;

/**
 * Represents a deletion in a genome.
 */
public class Deletion extends Variant {

  // The size of this deletion
  private long size;

  /**
   * Creates a new Deletion object.
   *
   * @param loc
   *          The location of this deletion (i.e. first deleted base in the
   *          chromosome)
   * @param size
   *          The size of this deletion
   * @param isIndel
   *          Whether or not this deletion is an indel (i.e. small insertion or
   *          deletion)
   */
  public Deletion(Location loc, long size, boolean isIndel) {
    super(loc, isIndel);

    // Parse arguments and throw errors
    if (size < 1) {
      throw new IllegalVariantException("Invalid size specified for deletion (non-positive value): " + size + "bp");
    } else if (loc.getCoord() + size - 1 > loc.getChr().getLength()) {
      throw new IllegalVariantException("Invalid size specified for deletion (too large for chromosome): "
          + loc.getChr().getName() + ":" + loc + " (" + size + "bp)");
    }

    // Initialize instance variable
    this.size = size;

    // Get the bases immediately before and after both breakpoints (up to MAX_MICROHOMOLOGY_WINDOW)
    String bp1fivePrimeBases = getFivePrimeBases(new Location(getChr(), getCoord() - 1));
    String bp1threePrimeBases = getThreePrimeBases(new Location(getChr(), getCoord() - 1));
    String bp2fivePrimeBases = getFivePrimeBases(getEndLoc());
    String bp2threePrimeBases = getThreePrimeBases(getEndLoc());

    // Find the number of bases that are identical on the 5' side of the two breakpoints
    int numFivePrimeIdenticalBases =
        getNumIdenticalBases(Genome.reverse(bp1fivePrimeBases), Genome.reverse(bp2fivePrimeBases));

    // Find the number of bases that are identical on the 3' side of the two breakpoints
    int numThreePrimeIdenticalBases = getNumIdenticalBases(bp1threePrimeBases, bp2threePrimeBases);

    // Define the range of possible breakpoint locations (location will represent the coordinate just before the break)
    Location bp1fivePrimeEdge = new Location(getChr(), getCoord() - 1 - numFivePrimeIdenticalBases);
    Location bp1threePrimeEdge = new Location(getChr(), getCoord() - 1 + numThreePrimeIdenticalBases);
    Location bp2fivePrimeEdge = new Location(getChr(), getEndCoord() - numFivePrimeIdenticalBases);
    Location bp2threePrimeEdge = new Location(getChr(), getEndCoord() + numThreePrimeIdenticalBases);

    // Set the breakpoints
    getBreakpoints().add(new Breakpoint(bp1fivePrimeEdge, bp1threePrimeEdge));
    getBreakpoints().add(new Breakpoint(bp2fivePrimeEdge, bp2threePrimeEdge));

  }

  /**
   * Returns the size of this deletion.
   *
   * @return The size of this deletion
   */
  public long getSize() {
    return size;
  }

  /**
   * Returns the end location of this deletion (i.e. last deleted base in the
   * chromosome).
   *
   * @return The end location of this deletion (i.e. last deleted base in the
   *         chromosome)
   */
  public Location getEndLoc() {
    return new Location(getChr(), getEndCoord());
  }

  /**
   * Returns the end coordinate of this deletion (i.e. last deleted base in the
   * chromosome).
   *
   * @return The end coordinate of this deletion (i.e. last deleted base in the
   *         chromosome)
   */
  public long getEndCoord() {
    return getCoord() + getSize() - 1;
  }

  /**
   * Returns the string of deleted bases for this deletion.
   *
   * @return The string of deleted bases for this deletion
   */
  public String getDeletedBases() {
    try {
      return getChr().getSequence(getCoord(), getEndCoord());
    } catch (IOException e) {
      throw new ShearRuntimeException("Error reading from the FASTA reference file", e);
    }
  }

  @Override
  public double getEstimatedReferenceDepth() {
    if (getBreakpoints().size() != 2) {
      throw new ShearRuntimeException("Exactly two breakpoints are expected for a deletion");
    }
    BreakpointReads bpReads1 = getBreakpoints().get(0).getBpReads();
    BreakpointReads bpReads2 = getBreakpoints().get(1).getBpReads();
    return ((double) bpReads1.getThreePrimeSpanningReads() + (double) bpReads2.getFivePrimeSpanningReads()) / 2d;
  }

  @Override
  public double getEstimatedVariantDepth() {
    if (getBreakpoints().size() != 2) {
      throw new ShearRuntimeException("Exactly two breakpoints are expected for a deletion");
    }
    BreakpointReads bpReads1 = getBreakpoints().get(0).getBpReads();
    BreakpointReads bpReads2 = getBreakpoints().get(1).getBpReads();
    return (double) bpReads1.getThreePrimeClippedReads() + (double) bpReads2.getFivePrimeClippedReads();
  }

  @Override
  public Deletion getCanonicalEquivalent() {
    return new Deletion(
        new Location(getChr(), getBreakpoints().get(0).getFivePrimeLoc().getCoord() + 1),
        getSize(),
        isIndel());
  }

  @Override
  public long getBreakpointMicrohomology() {
    long bp1microhomology =
        getBreakpoints().get(0).getThreePrimeLoc().getCoord() - getBreakpoints().get(0).getFivePrimeLoc().getCoord();
    long bp2microhomology =
        getBreakpoints().get(1).getThreePrimeLoc().getCoord() - getBreakpoints().get(1).getFivePrimeLoc().getCoord();
    if (bp1microhomology != bp2microhomology) {
      throw new ShearRuntimeException("Unexpected behavior while calculating breakpoint microhomology");
    }
    return bp1microhomology;
  }

  @Override
  public String getVariantType() {
    return "DEL";
  }

  @Override
  public String getVariantDetails() {
    return String.format("%s:%d-%d (%d bp)", getChr().getName(), getCoord(), getEndCoord(), getSize());
  }

  @Override
  public String getSdiString() {
    return formatSdiLine(getLoc(), getDeletedBases(), "-");
  }

  @Override
  public String getVariantIdString() {
    return super.getVariantIdString() + "\tDEL\t" + getSize();
  }

  @Override
  public boolean equals(Object obj) {
    if (super.equals(obj)) {
      if (obj == null || !(obj instanceof Deletion)) {
        return false;
      } else {
        Deletion other = (Deletion) obj;
        return getSize() == other.getSize();
      }
    } else {
      return false;
    }
  }

}
