/*
 * Sean Landman
 * Data Mining for Biomedical Informatics Group
 * University of Minnesota - Twin Cities
 */

package edu.umn.cs.kumarbio.shear.variants;

import java.io.IOException;

import edu.umn.cs.kumarbio.Genome;
import edu.umn.cs.kumarbio.shear.ShearRuntimeException;
import edu.umn.cs.kumarbio.shear.heterogeneity.Breakpoint;
import edu.umn.cs.kumarbio.shear.heterogeneity.BreakpointReads;

/**
 * Represents an insertion in a genome.
 */
public class Insertion extends Variant {

  // The string of inserted bases for this insertion
  private String insertedBases;

  /**
   * Creates a new Insertion object.
   *
   * @param loc
   *          The location of this insertion (i.e. position in the chromosome
   *          that will now appear directly after the inserted bases)
   * @param insertedBases
   *          The string of inserted bases for this insertion
   * @param isIndel
   *          Whether or not this insertion is an indel (i.e. small insertion or
   *          deletion)
   */
  public Insertion(Location loc, String insertedBases, boolean isIndel) {
    super(loc, isIndel);

    // Parse arguments and throw errors
    if (insertedBases == null) {
      throw new IllegalVariantException("Null inserted bases string specified for insertion");
    } else if (insertedBases.length() == 0) {
      throw new IllegalVariantException("Empty inserted bases string specified for insertion");
    } else if (!insertedBases.matches("^[ACGTNacgtn]+$")) {
      if (insertedBases.length() > 10) {
        throw new IllegalVariantException("Illegal inserted bases string specified for insertion: "
            + insertedBases.substring(0, 10) + "...");
      } else {
        throw new IllegalVariantException("Illegal inserted bases string specified for insertion: " + insertedBases);
      }
    }

    // Initialize instance variable
    this.insertedBases = insertedBases;

    // Get the bases immediately before and after the breakpoint (up to MAX_MICROHOMOLOGY_WINDOW)
    String bp1fivePrimeBases = getFivePrimeBases(new Location(getChr(), getCoord() - 1));
    String bp1threePrimeBases = getThreePrimeBases(new Location(getChr(), getCoord() - 1));

    // Find the number of bases that are identical between the end of the insertion and the 5' side of the breakpoint
    int numFivePrimeIdenticalBases = getNumIdenticalBases(bp1fivePrimeBases, Genome.reverse(insertedBases));

    // Find the number of bases that are identical between the start of the insertion and the 3' side of the breakpoint
    int numThreePrimeIdenticalBases = getNumIdenticalBases(bp1threePrimeBases, insertedBases);

    // Define the range of possible breakpoint locations (location will represent the coordinate just before the break)
    Location bp1fivePrimeEdge = new Location(getChr(), getCoord() - 1 - numFivePrimeIdenticalBases);
    Location bp1threePrimeEdge = new Location(getChr(), getCoord() - 1 + numThreePrimeIdenticalBases);

    // Set the breakpoints
    getBreakpoints().add(new Breakpoint(bp1fivePrimeEdge, bp1threePrimeEdge));

  }

  /**
   * Returns the string of inserted bases for this insertion.
   *
   * @return The string of inserted bases for this insertion
   */
  public String getInsertedBases() {
    return insertedBases;
  }

  /**
   * Returns the size of this insertion.
   *
   * @return The size of this insertion
   */
  public long getSize() {
    return getInsertedBases().length();
  }

  @Override
  public double getEstimatedReferenceDepth() {
    if (getBreakpoints().size() != 1) {
      throw new ShearRuntimeException("Exactly one breakpoint is expected for an insertion");
    }
    BreakpointReads bpReads1 = getBreakpoints().get(0).getBpReads();
    return (bpReads1.getFivePrimeSpanningReads() + bpReads1.getThreePrimeSpanningReads()) / 2d;
  }

  @Override
  public double getEstimatedVariantDepth() {
    if (getBreakpoints().size() != 1) {
      throw new ShearRuntimeException("Exactly one breakpoint is expected for an insertion");
    }
    BreakpointReads bpReads1 = getBreakpoints().get(0).getBpReads();
    return (bpReads1.getThreePrimeClippedReads() + (double) bpReads1.getFivePrimeClippedReads()) / 2d;
  }

  @Override
  public Insertion getCanonicalEquivalent() {
    long newBpFivePrimeShift = getCoord() - 1 - getBreakpoints().get(0).getFivePrimeLoc().getCoord();
    if (newBpFivePrimeShift > 0) {
      try {
        String newInsertedBases =
            getChr().getSequence(getCoord() - newBpFivePrimeShift, getCoord() - 1)
                + getInsertedBases().substring(0, getInsertedBases().length() - (int) newBpFivePrimeShift);
        return new Insertion(
            new Location(getChr(), getBreakpoints().get(0).getFivePrimeLoc().getCoord() + 1),
            newInsertedBases,
            isIndel());
      } catch (IOException e) {
        throw new ShearRuntimeException("I/O error occurred while reading from genome file");
      }
    } else {
      return new Insertion(getLoc(), getInsertedBases(), isIndel());
    }
  }

  @Override
  public long getBreakpointMicrohomology() {
    return getBreakpoints().get(0).getThreePrimeLoc().getCoord() - getBreakpoints().get(0).getFivePrimeLoc().getCoord();
  }

  @Override
  public String getVariantType() {
    return "INS";
  }

  @Override
  public String getVariantDetails() {
    return String.format("%s:%d (%d bp)", getChr().getName(), getCoord(), getSize());
  }

  @Override
  public String getSdiString() {
    return formatSdiLine(getLoc(), "-", getInsertedBases());
  }

  @Override
  public String getVariantIdString() {
    return super.getVariantIdString() + "\tINS\t" + getInsertedBases();
  }

  @Override
  public boolean equals(Object obj) {
    if (super.equals(obj)) {
      if (obj == null || !(obj instanceof Insertion)) {
        return false;
      } else {
        Insertion other = (Insertion) obj;
        return getInsertedBases().equals(other.getInsertedBases());
      }
    } else {
      return false;
    }
  }

}
