/*
 * Sean Landman
 * Data Mining for Biomedical Informatics Group
 * University of Minnesota - Twin Cities
 */

package edu.umn.cs.kumarbio.shear.variants;

import java.io.IOException;

import edu.umn.cs.kumarbio.Genome;
import edu.umn.cs.kumarbio.shear.ShearRuntimeException;
import edu.umn.cs.kumarbio.shear.heterogeneity.Breakpoint;
import edu.umn.cs.kumarbio.shear.heterogeneity.BreakpointReads;

/**
 * Represents an inversion in a genome.
 */
public class Inversion extends Variant {

  // The size of this inversion
  private long size;

  /**
   * Creates a new Inversion object.
   *
   * @param loc
   *          The location of this inversion (i.e. first affected base in the
   *          chromosome)
   * @param size
   *          The size of this inversion
   */
  public Inversion(Location loc, long size) {
    super(loc);

    // Parse arguments and throw errors
    if (size < 1) {
      throw new IllegalVariantException("Invalid size specified for inversion (non-positive value): " + size + "bp");
    } else if (loc.getCoord() + size - 1 > loc.getChr().getLength()) {
      throw new IllegalVariantException("Invalid size specified for inversion (too large for chromosome): "
          + loc.getChr().getName() + ":" + loc + " (" + size + "bp)");
    }

    // Initialize instance variable
    this.size = size;

    // Get the bases immediately before and after both breakpoints (up to MAX_MICROHOMOLOGY_WINDOW)
    String bp1fivePrimeBases = getFivePrimeBases(new Location(getChr(), getCoord() - 1));
    String bp1threePrimeBases = getThreePrimeBases(new Location(getChr(), getCoord() - 1));
    String bp2fivePrimeBases = getFivePrimeBases(getEndLoc());
    String bp2threePrimeBases = getThreePrimeBases(getEndLoc());

    // Find the number of bases that are identical on the outsides of the two breakpoints (i.e. 5' of first, 3' of second)
    // This compares forward strand bases at the first breakpoint against reverse strand bases at the second breakpoint
    int numOutsideIdenticalBases =
        getNumIdenticalBases(Genome.reverse(bp1fivePrimeBases), Genome.complement(bp2threePrimeBases));

    // Find the number of bases that are identical on the insides of the two breakpoints (i.e. 3' of first, 5' of second)
    // This compares forward strand bases at the first breakpoint against reverse strand bases at the second breakpoint
    int numInsideIdenticalBases = getNumIdenticalBases(bp1threePrimeBases, Genome.rc(bp2fivePrimeBases));

    // Define the range of possible breakpoint locations (location will represent the coordinate just before the break)
    Location bp1fivePrimeEdge = new Location(getChr(), getCoord() - 1 - numOutsideIdenticalBases);
    Location bp1threePrimeEdge = new Location(getChr(), getCoord() - 1 + numInsideIdenticalBases);
    Location bp2fivePrimeEdge = new Location(getChr(), getEndCoord() - numInsideIdenticalBases);
    Location bp2threePrimeEdge = new Location(getChr(), getEndCoord() + numOutsideIdenticalBases);

    // Set the breakpoints
    getBreakpoints().add(new Breakpoint(bp1fivePrimeEdge, bp1threePrimeEdge));
    getBreakpoints().add(new Breakpoint(bp2fivePrimeEdge, bp2threePrimeEdge));

  }

  /**
   * Returns the size of this inversion.
   *
   * @return The size of this inversion
   */
  public long getSize() {
    return size;
  }

  /**
   * Returns the end location of this inversion (i.e. last affected base in the
   * chromosome).
   *
   * @return The end location of this inversion (i.e. last affected base in the
   *         chromosome)
   */
  public Location getEndLoc() {
    return new Location(getChr(), getEndCoord());
  }

  /**
   * Returns the end coordinate of this inversion (i.e. last affected base in
   * the chromosome).
   *
   * @return The end coordinate of this inversion (i.e. last affected base in
   *         the chromosome)
   */
  public long getEndCoord() {
    return getCoord() + getSize() - 1;
  }

  /**
   * Returns the string of affected bases (before inverting) for this inversion.
   *
   * @return The string of affected bases (before inverting) for this inversion
   */
  public String getAffectedBases() {
    try {
      return getChr().getSequence(getCoord(), getEndCoord());
    } catch (IOException e) {
      throw new ShearRuntimeException("Error reading from the FASTA reference file", e);
    }
  }

  /**
   * Returns the string of inverted bases for this inversion.
   *
   * @return The string of inverted bases for this inversion
   */
  public String getInvertedBases() {
    return Genome.rc(getAffectedBases());
  }

  @Override
  public double getEstimatedReferenceDepth() {
    if (getBreakpoints().size() != 2) {
      throw new ShearRuntimeException("Exactly two breakpoints are expected for an inversion");
    }
    BreakpointReads bpReads1 = getBreakpoints().get(0).getBpReads();
    BreakpointReads bpReads2 = getBreakpoints().get(1).getBpReads();
    double bp1avgSpanningReads =
        ((double) bpReads1.getFivePrimeSpanningReads() + (double) bpReads1.getThreePrimeSpanningReads()) / 2d;
    double bp2avgSpanningReads =
        ((double) bpReads2.getFivePrimeSpanningReads() + (double) bpReads2.getThreePrimeSpanningReads()) / 2d;
    return (bp1avgSpanningReads + bp2avgSpanningReads) / 2d;
  }

  @Override
  public double getEstimatedVariantDepth() {
    if (getBreakpoints().size() != 2) {
      throw new ShearRuntimeException("Exactly two breakpoints are expected for an inversion");
    }
    BreakpointReads bpReads1 = getBreakpoints().get(0).getBpReads();
    BreakpointReads bpReads2 = getBreakpoints().get(1).getBpReads();
    return ((double) bpReads1.getFivePrimeClippedReads() + (double) bpReads1.getThreePrimeClippedReads()
        + bpReads2.getFivePrimeClippedReads() + bpReads2.getThreePrimeClippedReads()) / 2d;
  }

  @Override
  public Inversion getCanonicalEquivalent() {
    long sizeDifference = 2 * (getCoord() - 1 - getBreakpoints().get(0).getFivePrimeLoc().getCoord());
    return new Inversion(new Location(getChr(), getBreakpoints().get(0).getFivePrimeLoc().getCoord() + 1), getSize()
        + sizeDifference);
  }

  @Override
  public long getBreakpointMicrohomology() {
    long bp1microhomology =
        getBreakpoints().get(0).getThreePrimeLoc().getCoord() - getBreakpoints().get(0).getFivePrimeLoc().getCoord();
    long bp2microhomology =
        getBreakpoints().get(1).getThreePrimeLoc().getCoord() - getBreakpoints().get(1).getFivePrimeLoc().getCoord();
    if (bp1microhomology != bp2microhomology) {
      throw new ShearRuntimeException("Unexpected behavior while calculating breakpoint microhomology");
    }
    return bp1microhomology;
  }

  @Override
  public String getVariantType() {
    return "INV";
  }

  @Override
  public String getVariantDetails() {
    return String.format("%s:%d-%d (%d bp)", getChr().getName(), getCoord(), getEndCoord(), getSize());
  }

  @Override
  public String getSdiString() {
    return formatSdiLine(getLoc(), getAffectedBases(), getInvertedBases());
  }

  @Override
  public String getVariantIdString() {
    return super.getVariantIdString() + "\tINV\t" + getSize();
  }

  @Override
  public boolean equals(Object obj) {
    if (super.equals(obj)) {
      if (obj == null || !(obj instanceof Inversion)) {
        return false;
      } else {
        Inversion other = (Inversion) obj;
        return getSize() == other.getSize();
      }
    } else {
      return false;
    }
  }

}
