/*
 * Sean Landman
 * Data Mining for Biomedical Informatics Group
 * University of Minnesota - Twin Cities
 */

package edu.umn.cs.kumarbio.shear.variants;

import edu.umn.cs.kumarbio.Chromosome;
import edu.umn.cs.kumarbio.Genome;
import edu.umn.cs.kumarbio.shear.ShearRuntimeException;
import edu.umn.cs.kumarbio.shear.heterogeneity.Breakpoint;
import edu.umn.cs.kumarbio.shear.heterogeneity.BreakpointReads;

/**
 * Represents a translocation in a genome. This is represented by two locations
 * which are joined. The first location should refer to a base on the forward
 * strand. The second location specifies the base which is joined to the first
 * location, but this can refer to the base on the forward strand (3' to 5' end
 * join), to the 5' side of a base on the reverse strand joined to the 3' side
 * of the first location, or to the 3' side of a base on the reverse strand
 * joined to the 5' side of the first location.
 */
public class Translocation extends Variant {

  // The second location that this translocation involves
  private Location loc2;

  // The type of translocation join that this translocation involves
  private Type     transType;

  /**
   * This represents the possible types of translocation joins. The first
   * location in the Translocation class always refers to a base on the forward
   * strand. There are three possibilities for how this base is joined with the
   * second location:
   * <ul>
   * <li>for3toFor5: The 3' side of the forward base at the first loc is joined
   * to the 5' side of the forward base at the second loc</li>
   * <li>for3toRev5: The 3' side of the forward base at the first loc is joined
   * to the 5' side of the reverse base at the second loc</li>
   * <li>for5toRev3: The 5' side of the forward base at the first loc is joined
   * to the 3' side of the reverse base at the second loc</li>
   * </ul>
   */
  public enum Type {
    for3toFor5, for3toRev5, for5toRev3
  };

  /**
   * Create a new Translocation object.
   *
   * @param loc1
   *          The first location that this translocation involves
   * @param loc2
   *          The second location that this translocation involves
   * @param transType
   *          The type of translocation join that this translocation involves
   */
  public Translocation(Location loc1, Location loc2, Type transType) {
    super(loc1);

    // Parse arguments and throw errors
    if (loc2 == null) {
      throw new IllegalVariantException("Null second location specified for translocation");
    } else if (transType == null) {
      throw new IllegalVariantException("Null translocation type specified for translocation");
    } else if (loc1.getChr().getGenome() != loc2.getChr().getGenome()) {
      throw new IllegalVariantException(
          "Incompatable chromosomes specified for translocation (different Genome objects): " + loc1.getChr().getName()
              + " / " + loc2.getChr().getName());
    }

    // Initialize instance variables
    this.loc2 = loc2;
    this.transType = transType;

    // Range of possible breakpoint locations
    Location bp1fivePrimeEdge = null;
    Location bp1threePrimeEdge = null;
    Location bp2fivePrimeEdge = null;
    Location bp2threePrimeEdge = null;

    // Determine breakpoints for forward/3' to forward/5' join
    if (getTransType() == Type.for3toFor5) {

      // Get the forward strand bases immediately before and after both breakpoints (up to MAX_MICROHOMOLOGY_WINDOW)
      String bp1fivePrimeBases = getFivePrimeBases(getLoc());
      String bp1threePrimeBases = getThreePrimeBases(getLoc());
      String bp2fivePrimeBases = getFivePrimeBases(new Location(getSecondChr(), getSecondCoord() - 1));
      String bp2threePrimeBases = getThreePrimeBases(new Location(getSecondChr(), getSecondCoord() - 1));

      // Find the number of bases that are identical on the 5' side of the two breakpoints
      int numFivePrimeIdenticalBases =
          getNumIdenticalBases(Genome.reverse(bp1fivePrimeBases), Genome.reverse(bp2fivePrimeBases));

      // Find the number of bases that are identical on the 3' side of the two breakpoints
      int numThreePrimeIdenticalBases = getNumIdenticalBases(bp1threePrimeBases, bp2threePrimeBases);

      // Define the range of possible breakpoint locations (location will represent the coordinate just before the break)
      bp1fivePrimeEdge = new Location(getChr(), getCoord() - numFivePrimeIdenticalBases);
      bp1threePrimeEdge = new Location(getChr(), getCoord() + numThreePrimeIdenticalBases);
      bp2fivePrimeEdge = new Location(getSecondChr(), getSecondCoord() - 1 - numFivePrimeIdenticalBases);
      bp2threePrimeEdge = new Location(getSecondChr(), getSecondCoord() - 1 + numThreePrimeIdenticalBases);

      // Determine breakpoints for forward/3' to reverse/5' join
    } else if (getTransType() == Type.for3toRev5) {

      // Get the forward strand bases immediately before and after both breakpoints (up to MAX_MICROHOMOLOGY_WINDOW)
      String bp1fivePrimeBases = getFivePrimeBases(getLoc());
      String bp1threePrimeBases = getThreePrimeBases(getLoc());
      String bp2fivePrimeBases = getFivePrimeBases(getSecondLoc());
      String bp2threePrimeBases = getThreePrimeBases(getSecondLoc());

      // Find the number of bases that are identical on the 5' side of the two breakpoints
      // Note that the 5' side of the second breakpoint is actually "to the right" since it is on the reverse strand
      // Thus, for the second breakpoint, we look at the 3' side forward strand bases and take their complement
      int numFivePrimeIdenticalBases =
          getNumIdenticalBases(Genome.reverse(bp1fivePrimeBases), Genome.complement(bp2threePrimeBases));

      // Find the number of bases that are identical on the 3' side of the two breakpoints
      // Note that the 3' side of the second breakpoint is actually "to the left" since it is on the reverse strand
      // Thus, for the second breakpoint, we look at the 5' side forward strand bases and take their complement
      int numThreePrimeIdenticalBases = getNumIdenticalBases(bp1threePrimeBases, Genome.rc(bp2fivePrimeBases));

      // Define the range of possible breakpoint locations (location will represent the coordinate just before the break)
      // Note that for bp2, the 5' edge shifts by the number of 3' side bases identical between bp1/forward and bp2/reverse
      // Likewise the 3' edge for bp2 shifts by the number of 5' side bases identical between bp1/forward and bp2/reverse
      bp1fivePrimeEdge = new Location(getChr(), getCoord() - numFivePrimeIdenticalBases);
      bp1threePrimeEdge = new Location(getChr(), getCoord() + numThreePrimeIdenticalBases);
      bp2fivePrimeEdge = new Location(getSecondChr(), getSecondCoord() - numThreePrimeIdenticalBases);
      bp2threePrimeEdge = new Location(getSecondChr(), getSecondCoord() + numFivePrimeIdenticalBases);

      // Determine breakpoints for forward/5' to reverse/3' join
    } else if (getTransType() == Type.for5toRev3) {

      // Get the forward strand bases immediately before and after both breakpoints (up to MAX_MICROHOMOLOGY_WINDOW)
      String bp1fivePrimeBases = getFivePrimeBases(new Location(getChr(), getCoord() - 1));
      String bp1threePrimeBases = getThreePrimeBases(new Location(getChr(), getCoord() - 1));
      String bp2fivePrimeBases = getFivePrimeBases(new Location(getSecondChr(), getSecondCoord() - 1));
      String bp2threePrimeBases = getThreePrimeBases(new Location(getSecondChr(), getSecondCoord() - 1));

      // Find the number of bases that are identical on the 5' side of the two breakpoints
      // Note that the 5' side of the second breakpoint is actually "to the right" since it is on the reverse strand
      // Thus, for the second breakpoint, we look at the 3' side forward strand bases and take their complement
      int numFivePrimeIdenticalBases =
          getNumIdenticalBases(Genome.reverse(bp1fivePrimeBases), Genome.complement(bp2threePrimeBases));

      // Find the number of bases that are identical on the 3' side of the two breakpoints
      // Note that the 3' side of the second breakpoint is actually "to the left" since it is on the reverse strand
      // Thus, for the second breakpoint, we look at the 5' side forward strand bases and take their complement
      int numThreePrimeIdenticalBases = getNumIdenticalBases(bp1threePrimeBases, Genome.rc(bp2fivePrimeBases));

      // Define the range of possible breakpoint locations (location will represent the coordinate just before the break)
      // Note that for bp2, the 5' edge shifts by the number of 3' side bases identical between bp1/forward and bp2/reverse
      // Likewise the 3' edge for bp2 shifts by the number of 5' side bases identical between bp1/forward and bp2/reverse
      bp1fivePrimeEdge = new Location(getChr(), getCoord() - 1 - numFivePrimeIdenticalBases);
      bp1threePrimeEdge = new Location(getChr(), getCoord() - 1 + numThreePrimeIdenticalBases);
      bp2fivePrimeEdge = new Location(getSecondChr(), getSecondCoord() - 1 - numThreePrimeIdenticalBases);
      bp2threePrimeEdge = new Location(getSecondChr(), getSecondCoord() - 1 + numFivePrimeIdenticalBases);

      // Illegal translocation type
    } else {
      throw new ShearRuntimeException("Illegal translocation type specified");
    }

    // Set the breakpoints
    getBreakpoints().add(new Breakpoint(bp1fivePrimeEdge, bp1threePrimeEdge));
    getBreakpoints().add(new Breakpoint(bp2fivePrimeEdge, bp2threePrimeEdge));

  }

  /**
   * Returns the second chromosome that this translocation involves.
   *
   * @return The second chromosome that this translocation involves
   */
  public Chromosome getSecondChr() {
    return getSecondLoc().getChr();
  }

  /**
   * Returns the second coordinate that this translocation involves.
   *
   * @return The second coordinate that this translocation involves.
   */
  public long getSecondCoord() {
    return getSecondLoc().getCoord();
  }

  /**
   * Returns the second location that this translocation involves.
   *
   * @return The second location that this translocation involves
   */
  public Location getSecondLoc() {
    return loc2;
  }

  /**
   * Returns the type of translocation join that this translocation involves
   *
   * @return The type of translocation join that this translocation involves
   */
  public Type getTransType() {
    return transType;
  }

  @Override
  public double getEstimatedReferenceDepth() {
    if (getBreakpoints().size() != 2) {
      throw new ShearRuntimeException("Exactly two breakpoints are expected for a translocation");
    }
    BreakpointReads bpReads1 = getBreakpoints().get(0).getBpReads();
    BreakpointReads bpReads2 = getBreakpoints().get(1).getBpReads();
    switch (getTransType()) {
      case for3toFor5:
        return ((double) bpReads1.getThreePrimeSpanningReads() + (double) bpReads2.getFivePrimeSpanningReads()) / 2d;
      case for3toRev5:
        return ((double) bpReads1.getThreePrimeSpanningReads() + (double) bpReads2.getThreePrimeSpanningReads()) / 2d;
      case for5toRev3:
        return ((double) bpReads1.getFivePrimeSpanningReads() + (double) bpReads2.getFivePrimeSpanningReads()) / 2d;
      default:
        throw new ShearRuntimeException("Illegal translocation type");
    }
  }

  @Override
  public double getEstimatedVariantDepth() {
    if (getBreakpoints().size() != 2) {
      throw new ShearRuntimeException("Exactly two breakpoints are expected for a translocation");
    }
    BreakpointReads bpReads1 = getBreakpoints().get(0).getBpReads();
    BreakpointReads bpReads2 = getBreakpoints().get(1).getBpReads();
    switch (getTransType()) {
      case for3toFor5:
        return (double) bpReads1.getThreePrimeClippedReads() + (double) bpReads2.getFivePrimeClippedReads();
      case for3toRev5:
        return (double) bpReads1.getThreePrimeClippedReads() + (double) bpReads2.getThreePrimeClippedReads();
      case for5toRev3:
        return (double) bpReads1.getFivePrimeClippedReads() + (double) bpReads2.getFivePrimeClippedReads();
      default:
        throw new ShearRuntimeException("Illegal translocation type");
    }
  }

  @Override
  public Translocation getCanonicalEquivalent() {
    switch (getTransType()) {
      case for3toFor5:
        return new Translocation(
            new Location(getChr(), getBreakpoints().get(0).getFivePrimeLoc().getCoord()),
            new Location(getSecondChr(), getBreakpoints().get(1).getFivePrimeLoc().getCoord() + 1),
            getTransType());
      case for3toRev5:
        return new Translocation(
            new Location(getChr(), getBreakpoints().get(0).getFivePrimeLoc().getCoord()),
            new Location(getSecondChr(), getBreakpoints().get(1).getThreePrimeLoc().getCoord()),
            getTransType());
      case for5toRev3:
        return new Translocation(
            new Location(getChr(), getBreakpoints().get(0).getFivePrimeLoc().getCoord() + 1),
            new Location(getSecondChr(), getBreakpoints().get(1).getThreePrimeLoc().getCoord() + 1),
            getTransType());
      default:
        throw new ShearRuntimeException("Illegal translocation type");
    }
  }

  @Override
  public long getBreakpointMicrohomology() {
    long bp1microhomology =
        getBreakpoints().get(0).getThreePrimeLoc().getCoord() - getBreakpoints().get(0).getFivePrimeLoc().getCoord();
    long bp2microhomology =
        getBreakpoints().get(1).getThreePrimeLoc().getCoord() - getBreakpoints().get(1).getFivePrimeLoc().getCoord();
    if (bp1microhomology != bp2microhomology) {
      throw new ShearRuntimeException("Unexpected behavior while calculating breakpoint microhomology");
    }
    return bp1microhomology;
  }

  @Override
  public String getVariantType() {
    return "XLOC";
  }

  @Override
  public String getVariantDetails() {
    switch (getTransType()) {
      case for3toFor5:
        return String.format("%s (F strand / 3' end) <-> %s (F strand / 5' end)", getLoc(), getSecondLoc());
      case for3toRev5:
        return String.format("%s (F strand / 3' end) <-> %s (R strand / 5' end)", getLoc(), getSecondLoc());
      case for5toRev3:
        return String.format("%s (F strand / 5' end) <-> %s (R strand / 3' end)", getLoc(), getSecondLoc());
      default:
        throw new ShearRuntimeException("Illegal translocation type specified");
    }
  }

  @Override
  public String getSdiString() {
    return null;
  }

  @Override
  public String getVariantIdString() {
    return super.getVariantIdString() + "\tXLOC\t" + getSecondLoc() + "\t" + getTransType();
  }

  @Override
  public boolean equals(Object obj) {
    if (super.equals(obj)) {
      if (obj == null || !(obj instanceof Translocation)) {
        return false;
      } else {
        Translocation other = (Translocation) obj;
        return getSecondLoc().equals(other.getSecondLoc()) && getTransType().equals(other.getTransType());
      }
    } else {
      return false;
    }
  }

}
