/*
 * Sean Landman
 * Data Mining for Biomedical Informatics Group
 * University of Minnesota - Twin Cities
 */

package edu.umn.cs.kumarbio;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.text.ParseException;
import java.util.ArrayList;
import java.util.List;
import java.util.Scanner;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * Represents a genome (i.e. FASTA file). Constructor parses the corresponding
 * FASTA index file.
 */
public class Genome {

  // The path to the FASTA file for this genome
  private String                faFilePath;

  // RandomAccessFile object used to access the FASTA file
  private RandomAccessFile      faFileReader;

  // List of the chromsomes in this genome
  private ArrayList<Chromosome> chromosomes;

  /**
   * Constructs a new Genome object using the provided FASTA and FASTA index
   * (i.e. FAI) files. Will parse the FAI file for correctness, creating
   * Chromosome objects that use the FAI information to index into the FASTA
   * file for future operations. Throws a FileFoundFoundException for bad paths,
   * a ParseException for unexpected formats, or an IOException if there are any
   * problems accessing the FASTA or FAI files.
   *
   * @param faFilePath
   *          The path to the FASTA file
   * @param faiFilePath
   *          The path to the FAI file
   * @throws FileNotFoundException
   *           If bad path or missing file(s)
   * @throws ParseException
   *           If bad format in the FASTA or FAI files
   * @throws IOException
   *           If there are any problems accessing the FASTA or FAI files
   */
  public Genome(String faFilePath, String faiFilePath) throws FileNotFoundException, ParseException, IOException {

    // Set the faFilePath instanceVariable
    this.faFilePath = faFilePath;

    // Initialize RandomAccessFile object used to access the FASTA file
    faFileReader = new RandomAccessFile(faFilePath, "r");

    // Initialize list of the chromsomes in this genome
    chromosomes = new ArrayList<Chromosome>();

    // Variables to use for each iteration of parsing the FAI file
    String name;
    long length;
    long offset;
    int lineLength;
    int lineBytes;
    Matcher m;
    String seqHeaderLine;
    String[] tokens;
    long seqHeaderOffset = 0;
    long numFullLines = 0;

    // Keep track of line number in FAI file
    int lineCtr = 0;

    // Pattern to match for each FAI line
    Pattern p = Pattern.compile("^(\\S+)\\t(\\d+)\\t(\\d+)\\t(\\d+)\\t(\\d+)$");

    // Loop through FAI file
    Scanner sc = new Scanner(new File(faiFilePath));
    while (sc.hasNextLine()) {
      lineCtr++;

      // Verify that FAI line matches the pattern
      m = p.matcher(sc.nextLine());
      if (!m.matches()) {
        close();
        sc.close();
        throw new ParseException("Unexpected format in file <" + faiFilePath + "> at line " + lineCtr, lineCtr);
      }

      // Parse FAI line
      try {
        name = m.group(1);
        length = Long.parseLong(m.group(2));
        offset = Long.parseLong(m.group(3));
        lineLength = Integer.parseInt(m.group(4));
        lineBytes = Integer.parseInt(m.group(5));
      } catch (NumberFormatException e) {
        close();
        sc.close();
        throw new ParseException("Unexpected format in file <" + faiFilePath + "> at line " + lineCtr, lineCtr);
      }

      // Verify that each line has one terminal newline character
      if (lineBytes != lineLength + 1) {
        close();
        sc.close();
        throw new ParseException("Unexpected format in file <" + faiFilePath + "> at line " + lineCtr, lineCtr);
      }

      // Verify that chromosome name in the FAI and FASTA file match each other
      // and that the chromosome header line is where we expect it in the FASTA file
      faFileReader.seek(seqHeaderOffset);
      seqHeaderLine = faFileReader.readLine();
      tokens = seqHeaderLine.split("\\s");
      if (tokens[0].charAt(0) != '>' || !tokens[0].substring(1).equals(name)) {
        close();
        sc.close();
        throw new ParseException("Discrepenacy in FASTA/FAI files while parsing <" + faiFilePath + "> at line "
            + lineCtr, lineCtr);
      }

      // Update seqHeaderOffset
      // If the chromosome length is divisible by the lineLength, then use just offset + number of full lines * lineBytes
      // Otherwise use offset + number of full lines * lineBytes + leftover chars on last line + newline bytes
      numFullLines = length / lineLength;
      if (length % lineLength == 0) {
        seqHeaderOffset = offset + numFullLines * lineBytes;
      } else {
        seqHeaderOffset = offset + numFullLines * lineBytes + length % lineLength + (lineBytes - lineLength);
      }

      // Create and add the chromosome object
      if (!containsChromosome(name)) {
        chromosomes.add(new Chromosome(this, name, length, offset, lineLength, lineBytes));
      } else {
        sc.close();
        throw new IllegalArgumentException("A chromosome with the name of '" + name + "' already exists in this genome");
      }

    }

    // Close FAI input
    sc.close();

  }

  /**
   * Returns the path to the FASTA file for this genome.
   *
   * @return The path to the FASTA file for this genome
   */
  public String getFaFilePath() {
    return faFilePath;
  }

  /**
   * Returns a list of the chromosomes in this genome.
   *
   * @return A list of the chromosomes in this genome
   */
  public List<Chromosome> getChromosomeList() {
    return chromosomes;
  }

  /**
   * Returns a reference to the RandomAccessFile object used to access the FASTA
   * file.
   *
   * @return A reference to the RandomAccessFile object used to access the FASTA
   *         file
   */
  public RandomAccessFile getFaFileReader() {
    return faFileReader;
  }

  /**
   * Returns the length of this genome, taken by summing the lengths of all of
   * its chromosomes.
   *
   * @return The length of this genome
   */
  public long getGenomeLength() {
    long genomeLength = 0;
    for (Chromosome chr : chromosomes) {
      genomeLength += chr.getLength();
    }
    return genomeLength;
  }

  /**
   * Returns the length of the chromosome with the given name.
   *
   * @param chrName
   *          The name of the chromosome to get the sequence from
   * @return the length of the chromosome with the given name
   */
  public long getChromosomeLength(String chrName) {
    return getChromosome(chrName).getLength();
  }

  /**
   * Returns the sequence string starting at startLoc and ending at endLoc (both
   * inclusive) for the chromosome with the given name. Can only work for
   * sequences roughly the size of 2^31 (~2,147,000,000) base pairs long or
   * shorter.
   *
   * @param chrName
   *          The name of the chromosome to get the sequence from
   * @param startLoc
   *          The starting location of the desired sequence (inclusive)
   * @param endLoc
   *          The ending location of the desired sequence (inclusive)
   * @return The sequence string starting at startLoc and ending at endLoc (both
   *         inclusive) for the chromosome with the given name
   * @throws IOException
   *           If there are problems reading from the FASTA file
   */
  public String getSequence(String chrName, long startLoc, long endLoc) throws IOException {
    return getChromosome(chrName).getSequence(startLoc, endLoc);
  }

  /**
   * Returns the full sequence string for the chromosome with the given name.
   * Can only work for chromosomes roughly the size of 2^31 (~2,147,000,000)
   * base pairs long or shorter.
   *
   * @param chrName
   *          The name of the chromosome to get the sequence from
   * @return The full sequence string for the chromosome with the given name
   * @throws IOException
   *           If there are problems reading from the FASTA file
   */
  public String getFullSequence(String chrName) throws IOException {
    return getChromosome(chrName).getFullSequence();
  }

  /**
   * Returns true if the genome contains a chromosome with the given name, false
   * otherwise.
   *
   * @param chrName
   *          The name of the chromosome to check for
   * @return True if the genome contains a chromosome with the given name, false
   *         otherwise
   */
  public boolean containsChromosome(String chrName) {
    if (chrName == null) {
      throw new IllegalArgumentException("Null chromosome name specified");
    }
    for (Chromosome chr : chromosomes) {
      if (chr.getName().equals(chrName)) {
        return true;
      }
    }
    return false;
  }

  /**
   * Returns true if the genome contains the given chromosome, false otherwise.
   *
   * @param chr
   *          The chromosome to check for
   * @return True if the genome contains the given chromosome, false otherwise
   */
  public boolean containsChromosome(Chromosome chr) {
    if (chr == null) {
      throw new IllegalArgumentException("Null chromosome specified");
    }
    return chromosomes.contains(chr);
  }

  /**
   * Returns the chromosome specified by the given name.
   *
   * @param chrName
   *          The name of the chromosome to return
   * @return The specified chromosome
   */
  public Chromosome getChromosome(String chrName) {
    if (containsChromosome(chrName)) {
      Chromosome returnChr = null;
      for (Chromosome currentChr : chromosomes) {
        if (currentChr.getName().equals(chrName)) {
          if (returnChr == null) {
            returnChr = currentChr;
          } else {
            throw new IllegalArgumentException("Chromosome '" + chrName + "' exists multiple times in this genome");
          }
        }
      }
      return returnChr;
    } else {
      throw new IllegalArgumentException("Chromosome '" + chrName + "' is not present in this genome");
    }
  }

  /**
   * Closes the genome file by closing the underlying RandomAccessFile object.
   *
   * @throws IOException
   *           If an I/O error occurs with closing the RandomAccessFile object
   */
  public void close() throws IOException {
    faFileReader.close();
  }

  /**
   * Returns the reverse of a sequence. Input is a StringBuffer object. Output
   * is a String object.
   *
   * @param seq
   *          The input sequence as a StringBuffer object
   * @return The reverse of the input sequence
   */
  public static String reverse(StringBuffer seq) {
    return seq.reverse().toString();
  }

  /**
   * Returns the reverse of a sequence. Input is a String object. Output is a
   * String object.
   *
   * @param seq
   *          The input sequence as a String object
   * @return The reverse of the input sequence
   */
  public static String reverse(String seq) {
    return reverse(new StringBuffer(seq));
  }

  /**
   * Returns the reverse complement of a sequence. Input is a StringBuffer
   * object. Output is a String object.
   *
   * @param seq
   *          The input sequence as a StringBuffer object
   * @return The reverse complement of the input sequence
   */
  public static String rc(StringBuffer seq) {
    return reverse(complement(seq));
  }

  /**
   * Returns the reverse complement of a sequence. Input is a String object.
   * Output is a String object.
   *
   * @param seq
   *          The input sequence as a String object
   * @return The reverse complement of the input sequence
   */
  public static String rc(String seq) {
    return rc(new StringBuffer(seq));
  }

  /**
   * Returns the complement of a sequence. Input is a StringBuffer object.
   * Output is a String object.
   *
   * @param seq
   *          The input sequence as a StringBuffer object
   * @return The complement of the input sequence
   */
  public static String complement(StringBuffer seq) {
    StringBuffer complement = new StringBuffer();
    for (int i = 0; i < seq.length(); i++) {
      switch (seq.charAt(i)) {
        case 'A':
          complement.append('T');
          break;
        case 'C':
          complement.append('G');
          break;
        case 'G':
          complement.append('C');
          break;
        case 'T':
          complement.append('A');
          break;
        case 'a':
          complement.append('t');
          break;
        case 'c':
          complement.append('g');
          break;
        case 'g':
          complement.append('c');
          break;
        case 't':
          complement.append('a');
          break;
        default:
          complement.append(seq.charAt(i));
          break;
      }
    }
    return complement.toString();
  }

  /**
   * Returns the complement of a sequence. Input is a String object. Output is a
   * String object.
   *
   * @param seq
   *          The input sequence as a String object
   * @return The complement of the input sequence
   */
  public static String complement(String seq) {
    return complement(new StringBuffer(seq));
  }

}
